from django.test import Client
from django.urls import reverse

from pytest_django.asserts import assertContains

from digid_eherkenning.choices import ConfigTypes
from digid_eherkenning.models import ConfigCertificate


def test_admin_changelist(admin_client: Client, digid_certificate):
    url = reverse("admin:digid_eherkenning_configcertificate_changelist")
    digid_certificate.label = "DigiD certificate"
    digid_certificate.save()
    ConfigCertificate.objects.create(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    response = admin_client.get(url)

    assert response.status_code == 200
    assertContains(response, "DigiD certificate")



from collections.abc import Sequence

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField
from mozilla_django_oidc_db.fields import ClaimField, ClaimFieldDefault
from mozilla_django_oidc_db.typing import ClaimPath

from ...choices import AssuranceLevels
from .base import (
    BaseConfig,
    default_loa_choices,
    get_default_scopes_bsn,
    get_default_scopes_kvk,
)


class AuthorizeeMixin(models.Model):
    # XXX: this may require a value mapping, depending on what brokers return
    # XXX: this may require a fallback value, depending on what brokers return
    identifier_type_claim = ClaimField(
        verbose_name=_("identifier type claim"),
        # XXX: Anoigo specific default
        default=ClaimFieldDefault("namequalifier"),
        help_text=_(
            "Claim that specifies how the legal subject claim must be interpreted. "
            "The expected claim value is one of: "
            "'urn:etoegang:1.9:EntityConcernedID:KvKnr' or "
            "'urn:etoegang:1.9:EntityConcernedID:RSIN'."
        ),
    )
    # TODO: what if the claims for kvk/RSIN are different claims names?
    legal_subject_claim = ClaimField(
        verbose_name=_("company identifier claim"),
        default=ClaimFieldDefault("urn:etoegang:core:LegalSubjectID"),
        help_text=_(
            "Name of the claim holding the identifier of the authenticated company."
        ),
    )
    acting_subject_claim = ClaimField(
        verbose_name=_("acting subject identifier claim"),
        default=ClaimFieldDefault("urn:etoegang:core:ActingSubjectID"),
        help_text=_(
            "Name of the claim holding the (opaque) identifier of the user "
            "representing the authenticated company.."
        ),
    )
    branch_number_claim = ClaimField(
        verbose_name=_("branch number claim"),
        default=ClaimFieldDefault("urn:etoegang:1.9:ServiceRestriction:Vestigingsnr"),
        help_text=_(
            "Name of the claim holding the value of the branch number for the "
            "authenticated company, if such a restriction applies."
        ),
    )

    CLAIMS_CONFIGURATION = (
        {"field": "identifier_type_claim", "required": False},
        {"field": "legal_subject_claim", "required": True},
        {"field": "acting_subject_claim", "required": True},
        {"field": "branch_number_claim", "required": False},
    )

    class Meta:
        abstract = True

    @property
    def oidcdb_sensitive_claims(self) -> Sequence[ClaimPath]:
        return [
            self.legal_subject_claim,  # type: ignore
            self.branch_number_claim,  # type: ignore
        ]


@default_loa_choices(AssuranceLevels)
class EHerkenningConfig(AuthorizeeMixin, BaseConfig):
    """
    Configuration for eHerkenning authentication via OpenID connect.
    """

    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes_kvk,
        blank=True,
        help_text=_(
            "OpenID Connect scopes that are requested during login. "
            "These scopes are hardcoded and must be supported by the identity provider."
        ),
    )

    class Meta:
        verbose_name = _("OpenID Connect configuration for eHerkenning")

    @property
    def oidcdb_username_claim(self) -> ClaimPath:
        return self.legal_subject_claim


@default_loa_choices(AssuranceLevels)
class EHerkenningBewindvoeringConfig(AuthorizeeMixin, BaseConfig):
    # NOTE: Discussion with an employee from Anoigo states this will always be a BSN,
    # not an RSIN or CoC number.
    representee_claim = ClaimField(
        verbose_name=_("representee identifier claim"),
        # TODO: this is Anoigo, but could really be anything...
        default=ClaimFieldDefault("sel_uid"),
        help_text=_("Name of the claim holding the BSN of the represented person."),
    )

    mandate_service_id_claim = ClaimField(
        verbose_name=_("service ID claim"),
        default=ClaimFieldDefault("urn:etoegang:core:ServiceID"),
        help_text=_(
            "Name of the claim holding the service ID for which the company "
            "is authorized."
        ),
    )
    mandate_service_uuid_claim = ClaimField(
        verbose_name=_("service UUID claim"),
        default=ClaimFieldDefault("urn:etoegang:core:ServiceUUID"),
        help_text=_(
            "Name of the claim holding the service UUID for which the company "
            "is authorized."
        ),
    )

    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes_bsn,
        blank=True,
        help_text=_(
            "OpenID Connect scopes that are requested during login. "
            "These scopes are hardcoded and must be supported by the identity provider."
        ),
    )

    CLAIMS_CONFIGURATION = AuthorizeeMixin.CLAIMS_CONFIGURATION + (
        {"field": "representee_claim", "required": True},
        {"field": "mandate_service_id_claim", "required": True},
        {"field": "mandate_service_uuid_claim", "required": True},
    )

    class Meta:
        verbose_name = _("OpenID Connect configuration for eHerkenning Bewindvoering")

    @property
    def oidcdb_sensitive_claims(self) -> Sequence[ClaimPath]:
        base = super().oidcdb_sensitive_claims
        return base + [
            self.representee_claim,  # type: ignore
        ]



from django.contrib.auth.models import AbstractUser

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.typing import JSONObject

from .models.base import BaseConfig


class BaseBackend(OIDCAuthenticationBackend):
    def _check_candidate_backend(self) -> bool:
        suitable_model = issubclass(self.config_class, BaseConfig)
        return suitable_model and super()._check_candidate_backend()

    def update_user(self, user: AbstractUser, claims: JSONObject):
        # do nothing by default
        return user



from copy import deepcopy
from typing import TypedDict

from django.conf import settings
from django.db import models
from django.utils.functional import classproperty
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import JSONField
from mozilla_django_oidc_db.fields import ClaimField
from mozilla_django_oidc_db.models import OpenIDConnectConfigBase


def get_default_scopes_bsn():
    """
    Returns the default scopes to request for OpenID Connect logins for DigiD.
    """
    return ["openid", "bsn"]


def get_default_scopes_kvk():
    """
    Returns the default scopes to request for OpenID Connect logins for eHerkenning.
    """
    return ["openid", "kvk"]


def default_loa_choices(choicesCls: type[models.TextChoices]):
    def decorator(cls: type[BaseConfig]):
        # set the choices for the default_loa
        default_loa_field = cls._meta.get_field("default_loa")
        assert isinstance(default_loa_field, models.CharField)
        default_loa_field.choices = choicesCls.choices

        # specify the choices for the JSONField schema
        loa_mapping_field = cls._meta.get_field("loa_value_mapping")
        assert isinstance(loa_mapping_field, JSONField)
        new_schema = deepcopy(loa_mapping_field.schema)
        new_schema["items"]["properties"]["to"]["choices"] = [
            {"value": val, "title": label} for val, label in choicesCls.choices
        ]
        loa_mapping_field.schema = new_schema

        return cls

    return decorator


LOA_MAPPING_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["from", "to"],
        "properties": {
            "from": {
                "anyOf": [
                    {
                        "type": "string",
                        "title": _("String value"),
                    },
                    {
                        "type": "number",
                        "title": _("Number value"),
                    },
                ],
            },
            "to": {
                "type": "string",
            },
        },
        "additionalProperties": False,
    },
}


class ClaimConfiguration(TypedDict):
    field: str  # model field name
    required: bool


class BaseConfig(OpenIDConnectConfigBase):
    """
    Base configuration for DigiD/eHerkenning authentication via OpenID Connect.
    """

    loa_claim = ClaimField(
        verbose_name=_("LoA claim"),
        default=None,
        help_text=_(
            "Name of the claim holding the level of assurance. If left empty, it is "
            "assumed there is no LOA claim and the configured fallback value will be "
            "used."
        ),
        null=True,
        blank=True,
    )
    default_loa = models.CharField(
        _("default LOA"),
        max_length=100,
        blank=True,
        choices=tuple(),  # set dynamically via the default_loa_choices decorator
        help_text=_(
            "Fallback level of assurance, in case no claim value could be extracted."
        ),
    )

    loa_value_mapping = JSONField(
        _("loa mapping"),
        schema=LOA_MAPPING_SCHEMA,
        default=list,
        blank=True,
        help_text=_(
            "Level of assurance claim value mappings. Useful if the values in the LOA "
            "claim are proprietary, so you can translate them into their standardized "
            "identifiers."
        ),
    )

    CLAIMS_CONFIGURATION: tuple[ClaimConfiguration, ...]

    class Meta:
        abstract = True

    @classproperty
    def oidcdb_check_idp_availability(cls) -> bool:
        return True

    def get_callback_view(self):
        configured_setting = getattr(
            settings,
            "DIGID_EHERKENNING_OIDC_CALLBACK_VIEW",
            "digid_eherkenning.oidc.views.default_callback_view",
        )
        return import_string(configured_setting)



# Generated by Django 4.2.13 on 2024-07-18 11:08

from io import BytesIO
from pathlib import Path

from django.core.files import File
from django.db import migrations

from simple_certmanager.utils import (
    BadPassword,
    KeyIsNotEncrypted,
    decrypted_key_to_pem,
    load_pem_x509_private_key,
)


def decrypt_encrypted_keys(apps, _):
    DigiDConfiguration = apps.get_model("digid_eherkenning", "DigiDConfiguration")
    EherkenningConfiguration = apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )

    for model in (DigiDConfiguration, EherkenningConfiguration):
        config = model.objects.first()

        # check if there is something to decrypt in the first place
        if (
            # no configuration record exists
            config is None
            # key is not encrypted (or we don't have the passphrase to decrypt it)
            or not (pw := config.key_passphrase)
            or not (cert := config.certificate)
            # no private key file is uploaded
            or not (privkey := cert.private_key)
            # or the DB thinks there is a file but it's not actually there (anymore)
            or not privkey.storage.exists(privkey.name)
        ):
            continue

        with privkey.open("rb") as keyfile:
            try:
                key = load_pem_x509_private_key(keyfile.read(), password=pw)
            # key material + passphrase set up are not in a consistent state, this would
            # also affect runtime behaviour, but we opt to make our migration robust.
            except (KeyIsNotEncrypted, BadPassword):
                continue

        decrypted_data = decrypted_key_to_pem(key)
        # replace existing data
        existing_name = Path(privkey.name).name  # strip nested directories of the name
        privkey.save(name=existing_name, content=File(BytesIO(decrypted_data)))


class Migration(migrations.Migration):

    dependencies = [
        ("digid_eherkenning", "0008_update_loa_fields"),
    ]

    operations = [
        # we cannot reverse it since we may not know the encryption key anymore
        migrations.RunPython(decrypt_encrypted_keys, migrations.RunPython.noop),
    ]



class SAML2Error(Exception):
    pass


class eHerkenningError(SAML2Error):
    pass


class eHerkenningNoRSINError(eHerkenningError):
    pass


class CertificateProblem(Exception):
    def __init__(self, msg: str, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)
        self.message = msg



from io import BytesIO
from typing import Callable, Literal

from django.core.files import File
from django.db import IntegrityError

import pytest
from cryptography.hazmat.primitives.asymmetric import rsa
from simple_certmanager.constants import CertificateTypes
from simple_certmanager.test.certificate_generation import key_to_pem
from simple_certmanager.utils import load_pem_x509_private_key

from digid_eherkenning.choices import AssuranceLevels, ConfigTypes


@pytest.mark.django_db()
def test_fixing_misconfigured_eherkenning(migrator):
    old_state = migrator.apply_initial_migration(
        ("digid_eherkenning", "0003_digidconfiguration_artifact_resolve_content_type")
    )
    OldEherkenningConfiguration = old_state.apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )

    # misconfigured entry; attribute had default, but also blank=True
    OldEherkenningConfiguration(loa="").save()

    new_state = migrator.apply_tested_migration(
        ("digid_eherkenning", "0004_alter_eherkenningconfiguration_loa")
    )
    EherkenningConfiguration = new_state.apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )
    config = EherkenningConfiguration.objects.get()
    assert config.loa == AssuranceLevels.low_plus

    # impossible to misconfigure
    config.loa = ""
    with pytest.raises(IntegrityError):
        config.save()


@pytest.mark.django_db
def test_decrypt_private_keys_with_passphrase(
    temp_private_root, migrator, encrypted_keypair: tuple[bytes, bytes]
):
    old_state = migrator.apply_initial_migration(
        ("digid_eherkenning", "0008_update_loa_fields")
    )
    OldDigiDConfiguration = old_state.apps.get_model(
        "digid_eherkenning", "DigiDConfiguration"
    )
    Certificate = old_state.apps.get_model("simple_certmanager", "Certificate")
    key, cert = encrypted_keypair
    certificate = Certificate.objects.create(
        label="Test certificate",
        type=CertificateTypes.key_pair,
        public_certificate=File(BytesIO(cert), name="client_cert.pem"),
        private_key=File(BytesIO(key), name="client_key.pem"),
    )
    OldDigiDConfiguration.objects.create(
        certificate=certificate, key_passphrase="SUPERSECRET🔐"
    )

    new_state = migrator.apply_tested_migration(
        ("digid_eherkenning", "0009_decrypt_private_keys")
    )

    DigiDConfiguration = new_state.apps.get_model(
        "digid_eherkenning", "DigiDConfiguration"
    )
    config = DigiDConfiguration.objects.get()
    with config.certificate.private_key.open("rb") as privkey:
        try:
            load_pem_x509_private_key(privkey.read(), password=None)
        except Exception:
            pytest.fail("Expected private key to be decrypted.")


def _decryption_skip_cases_idfn(case):
    model_name, has_config, passphrase, encrypt_key, *_ = case
    return f"test_decryption_skip_cases({model_name=} {has_config=} {passphrase=} {encrypt_key=})"


@pytest.mark.parametrize(
    "case",
    [
        # EH has to work too - working setup with the correct key
        (
            "EHerkenningConfiguration",
            True,
            "SUPERSECRET🔐",
            True,
            lambda key, cert: {
                "public_certificate": File(BytesIO(cert), name="client_cert.pem"),
                "private_key": File(BytesIO(key), name="client_key.pem"),
            },
        ),
        # No config actually exists
        (
            "DigiDConfiguration",
            False,
            "",
            False,
            lambda *args: None,
        ),
        # Config exists, but no cert is configured (required for django-solo to
        # properly work)
        (
            "DigiDConfiguration",
            True,
            "foo",
            False,
            lambda *args: None,
        ),
        # Config exists, certificate only has certificate (no privkey)
        (
            "DigiDConfiguration",
            True,
            "foo",
            False,
            lambda key, cert: {
                "public_certificate": File(BytesIO(cert), name="client_cert.pem"),
                "private_key": "",
            },
        ),
        # Config exists, passphrase set, but key is not encrypted
        (
            "DigiDConfiguration",
            True,
            "foo",
            False,
            lambda key, cert: {
                "public_certificate": File(BytesIO(cert), name="client_cert.pem"),
                "private_key": File(BytesIO(key), name="client_key.pem"),
            },
        ),
        # Config exists, passphrase is wrong
        (
            "DigiDConfiguration",
            True,
            "foo",
            True,
            lambda key, cert: {
                "public_certificate": File(BytesIO(cert), name="client_cert.pem"),
                "private_key": File(BytesIO(key), name="client_key.pem"),
            },
        ),
        # Config exists, passphrase is missing
        (
            "DigiDConfiguration",
            True,
            "",
            True,
            lambda key, cert: {
                "public_certificate": File(BytesIO(cert), name="client_cert.pem"),
                "private_key": File(BytesIO(key), name="client_key.pem"),
            },
        ),
    ],
    ids=_decryption_skip_cases_idfn,
)
@pytest.mark.django_db
def test_decryption_migration_robustness(
    temp_private_root,
    migrator,
    leaf_keypair: tuple[rsa.RSAPrivateKey, bytes],
    encrypted_keypair: tuple[bytes, bytes],
    case: tuple[
        Literal["DigiDConfiguration", "EHerkenningConfiguration"],
        bool,
        str,
        bool,
        Callable[[bytes, bytes], dict | None],
    ],
):
    model_name, has_config, passphrase, encrypt_key, certificate_kwargs_callback = case
    old_state = migrator.apply_initial_migration(
        ("digid_eherkenning", "0008_update_loa_fields")
    )
    OldConfig = old_state.apps.get_model("digid_eherkenning", model_name)
    Certificate = old_state.apps.get_model("simple_certmanager", "Certificate")
    encrypted_key, cert = encrypted_keypair
    key = encrypted_key if encrypt_key else key_to_pem(leaf_keypair[0], passphrase="")

    certificate_kwargs = certificate_kwargs_callback(key, cert)
    certificate = (
        None
        if certificate_kwargs is None
        else Certificate.objects.create(
            label="Test certificate",
            type=CertificateTypes.key_pair,
            **certificate_kwargs,
        )
    )
    if has_config:
        OldConfig.objects.create(certificate=certificate, key_passphrase=passphrase)

    try:
        migrator.apply_tested_migration(
            ("digid_eherkenning", "0009_decrypt_private_keys")
        )
    except Exception:
        pytest.fail("Expected migration not to crash")


@pytest.mark.parametrize(
    "model_name,config_type",
    [
        ("DigidConfiguration", ConfigTypes.digid),
        ("EherkenningConfiguration", ConfigTypes.eherkenning),
    ],
)
def test_move_certificate(
    temp_private_root,
    migrator,
    leaf_keypair,
    model_name,
    config_type,
):
    key, cert_pem = leaf_keypair
    key_pem = key_to_pem(key, passphrase="")
    old_state = migrator.apply_initial_migration(
        (
            "digid_eherkenning",
            "0011_configcertificate_configcertificate_uniq_config_cert",
        )
    )
    OldConfig = old_state.apps.get_model("digid_eherkenning", model_name)
    Certificate = old_state.apps.get_model("simple_certmanager", "Certificate")
    certificate = Certificate.objects.create(
        label="Test certificate",
        type=CertificateTypes.key_pair,
        public_certificate=File(BytesIO(cert_pem), name="client_cert.pem"),
        private_key=File(BytesIO(key_pem), name="client_key.pem"),
    )
    OldConfig.objects.create(certificate=certificate)

    new_state = migrator.apply_tested_migration(
        ("digid_eherkenning", "0012_move_config_certificate")
    )

    ConfigCertificate = new_state.apps.get_model(
        "digid_eherkenning", "ConfigCertificate"
    )
    instance = ConfigCertificate.objects.get(certificate=certificate.pk)
    assert instance.config_type == config_type


def test_move_certificate_noop(migrator):
    migrator.apply_initial_migration(
        (
            "digid_eherkenning",
            "0011_configcertificate_configcertificate_uniq_config_cert",
        )
    )
    # there are no config records on a fresh install
    new_state = migrator.apply_tested_migration(
        ("digid_eherkenning", "0012_move_config_certificate")
    )

    ConfigCertificate = new_state.apps.get_model(
        "digid_eherkenning", "ConfigCertificate"
    )
    assert not ConfigCertificate.objects.exists()



import logging

from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
from django.http import HttpResponseRedirect

from mozilla_django_oidc_db.views import (
    OIDCAuthenticationCallbackView as BaseCallbackView,
    OIDCInit,
)

from .models import (
    DigiDConfig,
    DigiDMachtigenConfig,
    EHerkenningBewindvoeringConfig,
    EHerkenningConfig,
)

logger = logging.getLogger(__name__)


class OIDCAuthenticationCallbackView(BaseCallbackView):
    """
    Check if the 'created user' from the authentication backend needs to be logged in.

    If we only want to perform the claim processing, then no real user is expected to
    be returned from the authentication backend, and hence we also don't want to try
    to log in this dummy user (as in, set ``request.user`` to a django user
    instance).

    Note that we deliberately don't perform these changes in :meth:`get` (anymore),
    since we miss the upstream library changes/fixes when we make invasive changes.
    Instead, the authentication backend receives all the necessary information and *is*
    the right place to implement this logic.
    """

    expect_django_user: bool = True
    """
    Set to ``True`` if a Django user is expected to be created by the backend.

    The OIDC backend is used just to obtain the claims via OIDC, but doesn't always need
    to result in a real Django user record being created.
    """

    user: AbstractBaseUser | AnonymousUser  # set on succesful auth/code exchange

    def login_success(self):
        """
        Overridden to not actually log the user in, since setting the BSN/KVK/... in
        the session variables is all that matters.
        """
        assert self.user

        match (self.expect_django_user, self.user.pk):
            case (False, pk) if pk is not None:
                raise TypeError(
                    "A real Django user instance was returned from the authentication "
                    "backend. This is a configuration/programming mistake!"
                )
            case (True, None):
                raise TypeError(
                    "A fake Django user instance was returned from the authentication "
                    "backend. This is a configuration/programming mistake!"
                )

        # default behaviour which logs the user in *iff* we expect a real Django user
        # to be returned, otherwise ignore all that.
        if self.expect_django_user:
            return super().login_success()

        return HttpResponseRedirect(self.success_url)


digid_init = OIDCInit.as_view(config_class=DigiDConfig)
eh_init = OIDCInit.as_view(config_class=EHerkenningConfig)
digid_machtigen_init = OIDCInit.as_view(config_class=DigiDMachtigenConfig)
eh_bewindvoering_init = OIDCInit.as_view(config_class=EHerkenningBewindvoeringConfig)

default_callback_view = OIDCAuthenticationCallbackView.as_view()



from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from onelogin.saml2.constants import OneLogin_Saml2_Constants
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from privates.fields import PrivateMediaFileField
from simple_certmanager.models import Certificate
from solo.models import SingletonModel

from ..choices import (
    ConfigTypes,
    DigestAlgorithms,
    SignatureAlgorithms,
    XMLContentTypes,
)
from ..exceptions import CertificateProblem
from .certificates import ConfigCertificate


class BaseConfiguration(SingletonModel):
    idp_metadata_file = PrivateMediaFileField(
        _("identity provider metadata"),
        blank=True,
        help_text=_(
            "The metadata file of the identity provider. This is auto populated "
            "from the configured source URL."
        ),
    )
    idp_service_entity_id = models.CharField(
        _("identity provider service entity ID"),
        max_length=255,
        blank=True,
        help_text=_(
            "Example value: 'https://was-preprod1.digid.nl/saml/idp/metadata'. Note "
            "that this must match the 'entityID' attribute on the "
            "'md:EntityDescriptor' node found in the Identity Provider's metadata. "
            "This is auto populated from the configured source URL."
        ),
    )
    metadata_file_source = models.URLField(
        _("metadata file(XML) URL"),
        max_length=255,
        default="",
        help_text=_(
            "The URL-source where the XML metadata file can be retrieved from."
        ),
    )
    want_assertions_signed = models.BooleanField(
        _("want assertions signed"),
        default=True,
        help_text=_(
            "If True, the XML assertions need to be signed, otherwise the whole "
            "response needs to be signed."
        ),
        max_length=100,
    )
    want_assertions_encrypted = models.BooleanField(
        _("want assertions encrypted"),
        default=False,
        help_text=_("If True the XML assertions need to be encrypted."),
        max_length=100,
    )
    artifact_resolve_content_type = models.CharField(
        _("resolve artifact binding content type"),
        choices=XMLContentTypes.choices,
        default=XMLContentTypes.soap_xml,
        max_length=100,
        help_text=_(
            "'application/soap+xml' is considered legacy and modern brokers typically "
            "expect 'text/xml'."
        ),
    )
    signature_algorithm = models.CharField(
        _("signature algorithm"),
        blank=True,
        choices=SignatureAlgorithms.choices,
        default=SignatureAlgorithms.rsa_sha1,
        help_text=_(
            "Signature algorithm. Note that DSA_SHA1 and RSA_SHA1 are deprecated, but "
            "RSA_SHA1 is still the default value in the SAMLv2 standard. Warning: "
            "there are known issues with single-logout functionality if using anything "
            "other than SHA1 due to some hardcoded algorithm."
        ),
        max_length=100,
    )
    digest_algorithm = models.CharField(
        _("digest algorithm"),
        blank=True,
        choices=DigestAlgorithms.choices,
        default=DigestAlgorithms.sha1,
        help_text=_(
            "Digest algorithm. Note that SHA1 is deprecated, but still the default "
            "value in the SAMLv2 standard. Warning: "
            "there are known issues with single-logout functionality if using anything "
            "other than SHA1 due to some hardcoded algorithm."
        ),
        max_length=100,
    )
    entity_id = models.CharField(
        _("entity ID"), help_text=_("Service provider entity ID."), max_length=100
    )
    base_url = models.URLField(
        _("base URL"),
        help_text=_("Base URL of the application, without trailing slash."),
        max_length=100,
    )
    service_name = models.CharField(
        _("service name"),
        help_text=_("Name of the service you are providing."),
        max_length=100,
    )
    service_description = models.CharField(
        _("service description"),
        help_text=_("A description of the service you are providing."),
        max_length=100,
    )
    technical_contact_person_telephone = models.CharField(
        _("technical contact: phone number"),
        blank=True,
        help_text=_(
            "Telephone number of the technical person responsible for this "
            "DigiD/eHerkenning/eIDAS setup. For it to show up in the metata, you "
            "should also specify the email address."
        ),
        max_length=100,
    )
    technical_contact_person_email = models.CharField(
        _("technical contact: email"),
        blank=True,
        help_text=_(
            "Email address of the technical person responsible for this "
            "DigiD/eHerkenning/eIDAS setup. For it to show up in the metadata, you "
            "should also specify the phone number."
        ),
        max_length=100,
    )
    organization_url = models.URLField(
        _("organization URL"),
        blank=True,
        help_text=_(
            "URL of the organization providing the service for which "
            "DigiD/eHerkenning/eIDAS login is configured. For it to show up in the "
            "metadata, you should also specify the organization name."
        ),
        max_length=255,
    )
    organization_name = models.CharField(
        _("organization name"),
        blank=True,
        help_text=_(
            "URL of the organization providing the service for which "
            "DigiD/eHerkenning/eIDAS login is configured. For it to show up in the "
            "metadata, you should also specify the organization URL."
        ),
        max_length=100,
    )

    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._active_metadata_file_source = self.metadata_file_source

    def __str__(self):
        return force_str(self._meta.verbose_name)

    def populate_xml_fields(self, urls: dict[str, str], xml: bytes) -> None:
        """
        Populates the idp_metadata_file and idp_service_entity_id fields based on the
        fetched xml metadata
        """
        self.idp_service_entity_id = urls["entityId"]
        content = ContentFile(xml)
        self.idp_metadata_file.save("metadata.xml", content, save=False)

    def process_metadata_from_xml_source(self) -> tuple[dict[str, str], bytes]:
        """
        Parses the xml metadata

        :return a tuple of a dictionary with the useful urls and the xml bytes.
        """
        try:
            xml = OneLogin_Saml2_IdPMetadataParser.get_metadata(
                self.metadata_file_source
            )
            parsed_idp_metadata = OneLogin_Saml2_IdPMetadataParser.parse(
                xml,
                required_sso_binding=OneLogin_Saml2_Constants.BINDING_HTTP_POST,
                required_slo_binding=OneLogin_Saml2_Constants.BINDING_HTTP_POST,
            )
        #  python3-saml library does not use proper-namespaced exceptions
        except Exception as exc:
            raise ValidationError(
                _("Failed to parse the metadata, got error: {err}").format(err=str(exc))
            ) from exc

        if not (idp := parsed_idp_metadata.get("idp")):
            raise ValidationError(
                _(
                    "Could not find any identity provider information in the metadata at the provided URL."
                )
            )

        # sometimes the xml file contains urn instead of a url as an entity ID
        # use the provided url instead
        urls = {
            "entityId": idp.get("entityId"),
            "sso_url": idp.get("singleSignOnService", {}).get("url"),
            "slo_url": idp.get("singleLogoutService", {}).get("url"),
        }

        return (urls, xml)

    def save(self, *args, **kwargs):
        force_update = kwargs.pop("force_metadata_update", False)
        if value := self.metadata_file_source:
            has_changed = value != self._active_metadata_file_source
            if force_update or has_changed:
                urls, xml = self.process_metadata_from_xml_source()
                self.populate_xml_fields(urls, xml)
                self._active_metadata_file_source = value

        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]
        super().save(*args, **kwargs)

    def clean(self):
        super().clean()

        # require that a certificate is configured
        if not ConfigCertificate.objects.for_config(self).exists():
            raise ValidationError(
                _(
                    "You must prepare at least one certificate for the {verbose_name}."
                ).format(verbose_name=self._meta.verbose_name)
            )

    @classmethod
    def _as_config_type(cls) -> ConfigTypes:
        opts = cls._meta
        return ConfigTypes(f"{opts.app_label}.{opts.object_name}")

    def select_certificates(self) -> tuple[Certificate, Certificate | None]:
        try:
            current_cert, next_cert = ConfigCertificate.objects.for_config(
                self
            ).select_certificates()
        except ConfigCertificate.DoesNotExist as exc:
            raise CertificateProblem(
                "No (valid) certificate configured. The configuration needs a "
                "certificate with private key and public certificate."
            ) from exc
        else:
            # type checker shanigans, mostly
            assert isinstance(current_cert, Certificate)
            assert next_cert is None or isinstance(next_cert, Certificate)

        return current_cert, next_cert



# Generated by Django 4.2.13 on 2024-06-21 10:17

from django.db import migrations, models

import mozilla_django_oidc_db.fields


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning_oidc_generics",
            "0006_alter_digidconfig_oidc_rp_scopes_list_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="digidconfig",
            name="default_loa",
            field=models.CharField(
                blank=True,
                choices=[
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport",
                        "DigiD Basis",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
                        "DigiD Midden",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:Smartcard",
                        "DigiD Substantieel",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:SmartcardPKI",
                        "DigiD Hoog",
                    ),
                ],
                help_text="Fallback level of assurance, in case no claim value could be extracted.",
                max_length=100,
                verbose_name="default LOA",
            ),
        ),
        migrations.AddField(
            model_name="digidconfig",
            name="loa_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=None,
                help_text="Name of the claim holding the level of assurance. If left empty, it is assumed there is no LOA claim and the configured fallback value will be used.",
                null=True,
                size=None,
                verbose_name="LoA claim",
            ),
        ),
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="default_loa",
            field=models.CharField(
                blank=True,
                choices=[
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport",
                        "DigiD Basis",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
                        "DigiD Midden",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:Smartcard",
                        "DigiD Substantieel",
                    ),
                    (
                        "urn:oasis:names:tc:SAML:2.0:ac:classes:SmartcardPKI",
                        "DigiD Hoog",
                    ),
                ],
                help_text="Fallback level of assurance, in case no claim value could be extracted.",
                max_length=100,
                verbose_name="default LOA",
            ),
        ),
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="loa_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=None,
                help_text="Name of the claim holding the level of assurance. If left empty, it is assumed there is no LOA claim and the configured fallback value will be used.",
                null=True,
                size=None,
                verbose_name="LoA claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="default_loa",
            field=models.CharField(
                blank=True,
                choices=[
                    ("urn:etoegang:core:assurance-class:loa1", "Non existent (1)"),
                    ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                    ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                    ("urn:etoegang:core:assurance-class:loa3", "Substantial (3)"),
                    ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                ],
                help_text="Fallback level of assurance, in case no claim value could be extracted.",
                max_length=100,
                verbose_name="default LOA",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="loa_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=None,
                help_text="Name of the claim holding the level of assurance. If left empty, it is assumed there is no LOA claim and the configured fallback value will be used.",
                null=True,
                size=None,
                verbose_name="LoA claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="default_loa",
            field=models.CharField(
                blank=True,
                choices=[
                    ("urn:etoegang:core:assurance-class:loa1", "Non existent (1)"),
                    ("urn:etoegang:core:assurance-class:loa2", "Low (2)"),
                    ("urn:etoegang:core:assurance-class:loa2plus", "Low (2+)"),
                    ("urn:etoegang:core:assurance-class:loa3", "Substantial (3)"),
                    ("urn:etoegang:core:assurance-class:loa4", "High (4)"),
                ],
                help_text="Fallback level of assurance, in case no claim value could be extracted.",
                max_length=100,
                verbose_name="default LOA",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="loa_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                blank=True,
                default=None,
                help_text="Name of the claim holding the level of assurance. If left empty, it is assumed there is no LOA claim and the configured fallback value will be used.",
                null=True,
                size=None,
                verbose_name="LoA claim",
            ),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class OIDCAppConfig(AppConfig):
    name = "digid_eherkenning.oidc"
    verbose_name = _("DigiD & eHerkenning via OpenID Connect")
    # can't change this label because of existing migrations in Open Forms/Open Inwoner
    label = "digid_eherkenning_oidc_generics"
    default_auto_field = "django.db.models.AutoField"



# Generated by Django 4.2.13 on 2024-06-11 13:43

from typing import Any, Callable

from django.conf import settings
from django.core.cache import caches
from django.db import migrations, models, transaction

import mozilla_django_oidc_db.fields


def flush_cache():
    cache_name = getattr(settings, "SOLO_CACHE", None)
    if not cache_name:
        return
    caches[cache_name].clear()


def operation_factory(model: str, mappings: dict[str, str]) -> migrations.RunPython:

    def _action_factory(transformer: Callable[[Any], None]):
        def _run_python_action(apps, _) -> None:
            ConfigModel = apps.get_model("digid_eherkenning_oidc_generics", model)

            # Solo model, so there's only ever one instance
            config = ConfigModel.objects.first()
            if config is None:
                return

            transformer(config)

            config.save()
            transaction.on_commit(flush_cache)

        return _run_python_action

    @_action_factory
    def forward(instance) -> None:
        for old, new in mappings.items():
            new_value = getattr(instance, old).split(".")
            setattr(instance, new, new_value)

    @_action_factory
    def reverse(instance) -> None:
        for old, new in mappings.items():
            old_value = ".".join(getattr(instance, new))
            setattr(instance, old, old_value)

    return migrations.RunPython(forward, reverse)


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning_oidc_generics",
            "0003_rename_openidconnectpublicconfig_digidconfig_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="digidconfig",
            name="bsn_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault("bsn"),
                help_text="Name of the claim holding the authenticated user's BSN.",
                size=None,
                verbose_name="bsn claim",
            ),
        ),
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="authorizee_bsn_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:nl-eid-gdi:1.0:ActingSubjectID"
                ),
                help_text="Name of the claim holding the BSN of the authorized user.",
                size=None,
                verbose_name="authorizee bsn claim",
            ),
        ),
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="representee_bsn_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:nl-eid-gdi:1.0:LegalSubjectID"
                ),
                help_text="Name of the claim holding the BSN of the represented user.",
                size=None,
                verbose_name="representee bsn claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="identifier_type_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "namequalifier"
                ),
                help_text="Claim that specifies how the legal subject claim must be interpreted. The expected claim value is one of: 'urn:etoegang:1.9:EntityConcernedID:KvKnr' or 'urn:etoegang:1.9:EntityConcernedID:RSIN'.",
                size=None,
                verbose_name="identifier type claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="legal_subject_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:LegalSubjectID"
                ),
                help_text="Name of the claim holding the identifier of the authenticated company.",
                size=None,
                verbose_name="company identifier claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="representee_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault("sel_uid"),
                help_text="Name of the claim holding the BSN of the represented person.",
                size=None,
                verbose_name="representee identifier claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="identifier_type_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "namequalifier"
                ),
                help_text="Claim that specifies how the legal subject claim must be interpreted. The expected claim value is one of: 'urn:etoegang:1.9:EntityConcernedID:KvKnr' or 'urn:etoegang:1.9:EntityConcernedID:RSIN'.",
                size=None,
                verbose_name="identifier type claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="legal_subject_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:LegalSubjectID"
                ),
                help_text="Name of the claim holding the identifier of the authenticated company.",
                size=None,
                verbose_name="company identifier claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="acting_subject_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:ActingSubjectID"
                ),
                help_text="Name of the claim holding the (opaque) identifier of the user representing the authenticated company..",
                size=None,
                verbose_name="acting subject identifier claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="acting_subject_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:ActingSubjectID"
                ),
                help_text="Name of the claim holding the (opaque) identifier of the user representing the authenticated company..",
                size=None,
                verbose_name="acting subject identifier claim",
            ),
        ),
        operation_factory(
            "DigiDConfig",
            mappings={
                "identifier_claim_name": "bsn_claim",
            },
        ),
        operation_factory(
            "DigiDMachtigenConfig",
            mappings={
                "vertegenwoordigde_claim_name": "representee_bsn_claim",
                "gemachtigde_claim_name": "authorizee_bsn_claim",
            },
        ),
        operation_factory(
            "EHerkenningConfig",
            mappings={
                "identifier_claim_name": "legal_subject_claim",
            },
        ),
        operation_factory(
            "EHerkenningBewindvoeringConfig",
            mappings={
                "vertegenwoordigde_company_claim_name": "representee_claim",
                "gemachtigde_person_claim_name": "acting_subject_claim",
            },
        ),
        migrations.RemoveField(
            model_name="digidconfig",
            name="identifier_claim_name",
        ),
        migrations.RemoveField(
            model_name="digidmachtigenconfig",
            name="gemachtigde_claim_name",
        ),
        migrations.RemoveField(
            model_name="digidmachtigenconfig",
            name="vertegenwoordigde_claim_name",
        ),
        migrations.RemoveField(
            model_name="eherkenningbewindvoeringconfig",
            name="gemachtigde_person_claim_name",
        ),
        migrations.RemoveField(
            model_name="eherkenningbewindvoeringconfig",
            name="vertegenwoordigde_company_claim_name",
        ),
        migrations.RemoveField(
            model_name="eherkenningconfig",
            name="identifier_claim_name",
        ),
    ]



# Generated by Django 4.2.13 on 2024-06-21 10:36

from django.db import migrations

import django_jsonform.models.fields


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning_oidc_generics",
            "0007_digidconfig_default_loa_digidconfig_loa_claim_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="digidconfig",
            name="loa_value_mapping",
            field=django_jsonform.models.fields.JSONField(
                default=list,
                blank=True,
                help_text="Level of assurance claim value mappings. Useful if the values in the LOA claim are proprietary, so you can translate them into their standardized identifiers.",
                verbose_name="loa mapping",
            ),
        ),
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="loa_value_mapping",
            field=django_jsonform.models.fields.JSONField(
                default=list,
                blank=True,
                help_text="Level of assurance claim value mappings. Useful if the values in the LOA claim are proprietary, so you can translate them into their standardized identifiers.",
                verbose_name="loa mapping",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="loa_value_mapping",
            field=django_jsonform.models.fields.JSONField(
                default=list,
                blank=True,
                help_text="Level of assurance claim value mappings. Useful if the values in the LOA claim are proprietary, so you can translate them into their standardized identifiers.",
                verbose_name="loa mapping",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="loa_value_mapping",
            field=django_jsonform.models.fields.JSONField(
                default=list,
                blank=True,
                help_text="Level of assurance claim value mappings. Useful if the values in the LOA claim are proprietary, so you can translate them into their standardized identifiers.",
                verbose_name="loa mapping",
            ),
        ),
    ]



"""
Functionality to relate one or more certificates to a SAMLv2 configuration.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, TypeAlias

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from simple_certmanager.constants import CertificateTypes
from simple_certmanager.models import Certificate

from ..choices import ConfigTypes

if TYPE_CHECKING:
    from .digid import DigidConfiguration
    from .eherkenning import EherkenningConfiguration

logger = logging.getLogger(__name__)

_AnyDigiD: TypeAlias = "type[DigidConfiguration] | DigidConfiguration"
_AnyEH: TypeAlias = "type[EherkenningConfiguration] | EherkenningConfiguration"


class ConfigCertificateQuerySet(models.QuerySet["ConfigCertificate"]):
    def for_config(self, config: _AnyDigiD | _AnyEH):
        config_type = config._as_config_type()
        return self.filter(config_type=config_type)

    def select_certificates(self) -> tuple[Certificate, Certificate | None]:
        """
        Select the best candidates for the current and next certificate.

        The certificates are used for signing authentication requests (and metadata)
        itself, and the possible certificates that perform signing are included in the
        metadata. For zero-downtime/gradual replacement, a current and next certificate
        can be provided (this is a limitation in python3-saml).

        We look for the current certificate and the next with the following algorithm:

        * order candidates by valid_from, so we favour existing/the oldest keypairs
        * order candidates by expiry date, so if they have an identical valid_from, we
          favour the one that will expiry first (the other one(s) automatically become
          the next certificate
        * discard any candidates that do not meet our key pair requirements, ignoring
          valid_from/until

        To determine the current certificate:

        * discard candidates that are not valid yet
        * discard candiates that are not valid anymore

        If no candidate matches, we raise a DoesNotExist exception.

        If a candidate is found, we select the next certificate according to:

        * must be valid_from >= current_certificate.valid_from
        * must not be expired
        """
        # XXX: check if this has a big performance impact because we extract the
        # valid_from/until by loading the certificate files!

        qs = self.filter(certificate__type=CertificateTypes.key_pair).iterator()
        # first pass - filter out anything that is not usable for SAML flows (
        # discarding broken/invalid configurations)
        candidates = [
            candidate
            for candidate in qs
            if candidate._meets_requirements_to_be_used_for_saml()
        ]
        # sort them - we now know that we can safely access the valid_from and
        # expiry_date attributes
        candidates = sorted(
            candidates,
            key=lambda c: (c.certificate.valid_from, c.certificate.expiry_date),
        )

        # figure out which certificate is our current certificate
        current_cert: Certificate | None = None
        next_cert: Certificate | None = None

        # loop only once, so that we are certain next_cert's validity is *after* current
        # cert.
        for candidate in candidates:
            certificate: Certificate = candidate.certificate
            match (current_cert, next_cert):
                case (None, None) if candidate.is_ready_for_authn_requests:
                    current_cert = certificate
                    continue  # the same candidate cannot both be current and next
                case (Certificate(), None) if certificate.expiry_date > timezone.now():
                    next_cert = certificate
                    break  # we found both current and next
        else:
            logger.debug("Could not determine a next certificate")

        if current_cert is None:
            raise self.model.DoesNotExist(
                "Could not find a suitable current certificate"
            )

        return current_cert, next_cert


class ConfigCertificateManager(models.Manager.from_queryset(ConfigCertificateQuerySet)):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("certificate")


class ConfigCertificate(models.Model):
    """
    Tie a particular certificate to a configuration model.
    """

    config_type = models.CharField(
        _("config type"),
        max_length=100,
        choices=ConfigTypes.choices,
    )
    certificate = models.ForeignKey(
        Certificate,
        on_delete=models.PROTECT,
        # Careful! This does not give any guarantees, you can select a valid certificate
        # and then make the certificate instance itself invalid, and end up with a
        # cert-only configuration.
        limit_choices_to={"type": CertificateTypes.key_pair},
        verbose_name=_("certificate"),
        help_text=_(
            "Certificate that may be used by the specified configuration. The best "
            "matching candidate will automatically be selected by the configuration."
        ),
    )

    objects = ConfigCertificateManager()

    class Meta:
        verbose_name = _("DigiD/eHerkenning certificate")
        verbose_name_plural = _("DigiD/eHerkenning certificates")
        constraints = [
            models.UniqueConstraint(
                name="uniq_config_cert",
                fields=("config_type", "certificate"),
                violation_error_message=_(
                    "This configuration and certificate combination already exists."
                ),
            ),
        ]

    def __str__(self):
        config_type = self.get_config_type_display()  # type: ignore
        _cert = self.certificate if self.certificate_id else None  # type: ignore
        certificate = str(_cert) if _cert else _("(no certificate selected)")
        return f"{config_type}: {certificate}"

    def _meets_requirements_to_be_used_for_saml(self) -> bool:
        try:
            _certificate: Certificate = self.certificate
        except Certificate.DoesNotExist:
            return False

        if _certificate.type != CertificateTypes.key_pair:
            return False

        if not (privkey := _certificate.private_key) or not privkey.storage.exists(
            privkey.name
        ):
            return False

        # Try loading it with cryptography
        try:
            _certificate.certificate
            valid_pair = _certificate.is_valid_key_pair()
        except (FileNotFoundError, ValueError) as exc:
            logger.info(
                "Could not introspect certificate validity",
                exc_info=exc,
                extra={"certificate_pk": _certificate.pk},
            )
            return False
        else:
            if not valid_pair:
                return False

        return True

    @property
    def is_ready_for_authn_requests(self) -> bool:
        """
        Introspect the certificate to determine if it's a candidate for authn requests.
        """
        if not self._meets_requirements_to_be_used_for_saml():
            return False

        _certificate: Certificate = self.certificate
        valid_from, expiry_date = _certificate.valid_from, _certificate.expiry_date

        now = timezone.now()
        if not (valid_from <= now <= expiry_date):
            return False

        return True



# Generated by Django 4.2.13 on 2024-07-18 14:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("digid_eherkenning", "0009_decrypt_private_keys"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="digidconfiguration",
            name="key_passphrase",
        ),
        migrations.RemoveField(
            model_name="eherkenningconfiguration",
            name="key_passphrase",
        ),
    ]



from .base import BaseConfig, get_default_scopes_bsn, get_default_scopes_kvk
from .digid import DigiDConfig, DigiDMachtigenConfig
from .eherkenning import EHerkenningBewindvoeringConfig, EHerkenningConfig

__all__ = [
    "get_default_scopes_bsn",
    "get_default_scopes_kvk",
    "BaseConfig",
    "DigiDConfig",
    "DigiDMachtigenConfig",
    "EHerkenningConfig",
    "EHerkenningBewindvoeringConfig",
]



# Generated by Django 4.2.13 on 2024-05-24 16:04

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("digid_eherkenning_oidc_generics", "0002_auto_20240207_1546"),
    ]

    operations = [
        migrations.RenameModel(
            old_name="OpenIDConnectPublicConfig",
            new_name="DigiDConfig",
        ),
        migrations.RenameModel(
            old_name="OpenIDConnectDigiDMachtigenConfig",
            new_name="DigiDMachtigenConfig",
        ),
        migrations.RenameModel(
            old_name="OpenIDConnectEHerkenningBewindvoeringConfig",
            new_name="EHerkenningBewindvoeringConfig",
        ),
        migrations.RenameModel(
            old_name="OpenIDConnectEHerkenningConfig",
            new_name="EHerkenningConfig",
        ),
    ]



from collections.abc import Sequence

from django.db import models
from django.utils.translation import gettext_lazy as _

from django_jsonform.models.fields import ArrayField
from mozilla_django_oidc_db.fields import ClaimField, ClaimFieldDefault
from mozilla_django_oidc_db.typing import ClaimPath

from ...choices import DigiDAssuranceLevels
from .base import BaseConfig, default_loa_choices, get_default_scopes_bsn


@default_loa_choices(DigiDAssuranceLevels)
class DigiDConfig(BaseConfig):
    """
    Configuration for DigiD authentication via OpenID connect
    """

    bsn_claim = ClaimField(
        verbose_name=_("bsn claim"),
        default=ClaimFieldDefault("bsn"),
        help_text=_("Name of the claim holding the authenticated user's BSN."),
    )
    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes_bsn,
        blank=True,
        help_text=_(
            "OpenID Connect scopes that are requested during login. "
            "These scopes are hardcoded and must be supported by the identity provider."
        ),
    )

    CLAIMS_CONFIGURATION = ({"field": "bsn_claim", "required": True},)

    class Meta:
        verbose_name = _("OpenID Connect configuration for DigiD")

    @property
    def oidcdb_username_claim(self) -> ClaimPath:
        return self.bsn_claim


@default_loa_choices(DigiDAssuranceLevels)
class DigiDMachtigenConfig(BaseConfig):
    # TODO: these default claim names don't appear to be part of any standard...
    representee_bsn_claim = ClaimField(
        verbose_name=_("representee bsn claim"),
        default=ClaimFieldDefault("urn:nl-eid-gdi:1.0:LegalSubjectID"),
        help_text=_("Name of the claim holding the BSN of the represented user."),
    )
    authorizee_bsn_claim = ClaimField(
        verbose_name=_("authorizee bsn claim"),
        default=ClaimFieldDefault("urn:nl-eid-gdi:1.0:ActingSubjectID"),
        help_text=_("Name of the claim holding the BSN of the authorized user."),
    )
    mandate_service_id_claim = ClaimField(
        verbose_name=_("service ID claim"),
        default=ClaimFieldDefault("urn:nl-eid-gdi:1.0:ServiceUUID"),
        help_text=_(
            "Name of the claim holding the service UUID for which the acting subject "
            "is authorized."
        ),
    )

    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes_bsn,
        blank=True,
        help_text=_(
            "OpenID Connect scopes that are requested during login. "
            "These scopes are hardcoded and must be supported by the identity provider."
        ),
    )

    CLAIMS_CONFIGURATION = (
        {"field": "representee_bsn_claim", "required": True},
        {"field": "authorizee_bsn_claim", "required": True},
        {"field": "mandate_service_id_claim", "required": True},
    )

    class Meta:
        verbose_name = _("OpenID Connect configuration for DigiD Machtigen")

    @property
    def oidcdb_sensitive_claims(self) -> Sequence[ClaimPath]:
        return [
            self.representee_bsn_claim,  # type: ignore
            self.authorizee_bsn_claim,  # type: ignore
        ]



from django.test import Client
from django.urls import reverse
from django.utils.translation import gettext as _

import pytest
from pytest_django.asserts import assertContains

from digid_eherkenning.models import DigidConfiguration, EherkenningConfiguration


@pytest.mark.django_db
def test_digid_configuration_admin_certificates_link(
    admin_client: Client,
    digid_config: DigidConfiguration,
):
    url = reverse("admin:digid_eherkenning_digidconfiguration_change", args=(1,))

    response = admin_client.get(url)

    assert response.status_code == 200
    assertContains(response, _("certificates"))
    assertContains(response, _("Manage ({count})").format(count=1))


@pytest.mark.django_db
def test_eherkenning_configuration_admin_certificates_link(
    admin_client: Client,
    eherkenning_config: EherkenningConfiguration,
):
    url = reverse("admin:digid_eherkenning_eherkenningconfiguration_change", args=(1,))

    response = admin_client.get(url)

    assert response.status_code == 200
    assertContains(response, _("certificates"))
    assertContains(response, _("Manage ({count})").format(count=1))



import binascii
from base64 import b64encode
from io import BytesIO
from typing import Union, no_type_check
from uuid import uuid4

from django.urls import reverse
from django.utils import timezone

from cryptography.hazmat.primitives import serialization
from cryptography.x509 import load_pem_x509_certificate
from lxml.builder import ElementMaker
from lxml.etree import Element, tostring
from onelogin.saml2.settings import OneLogin_Saml2_Settings

from ..models import EherkenningConfiguration
from ..settings import EHERKENNING_DS_XSD
from ..types import EHerkenningConfig, EHerkenningSAMLConfig, ServiceConfig
from ..utils import validate_xml
from .base import BaseSaml2Client, get_service_description, get_service_name

namespaces = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
    "esc": "urn:etoegang:1.13:service-catalog",
    "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
}
ESC = ElementMaker(namespace=namespaces["esc"], nsmap=namespaces)
DS = ElementMaker(namespace=namespaces["ds"], nsmap=namespaces)
SAML = ElementMaker(namespace=namespaces["saml"], nsmap=namespaces)
MD = ElementMaker(namespace=namespaces["md"], nsmap=namespaces)

xml_nl_lang = {"{http://www.w3.org/XML/1998/namespace}lang": "nl"}


def generate_dienst_catalogus_metadata(eherkenning_config=None):
    eherkenning_config = eherkenning_config or EherkenningConfiguration.get_solo()
    settings: EHerkenningConfig = eherkenning_config.as_dict()
    # ensure that the single language strings are output in both nl and en
    for service in settings["services"]:
        name = service["service_name"]
        service["service_name"] = {"nl": name, "en": name}

        description = service["service_description"]
        service["service_description"] = {"nl": description, "en": description}

        privacy_policy_url = service["privacy_policy_url"]
        service["privacy_policy_url"] = {"nl": privacy_policy_url}

        service_description_url = service["service_description_url"]
        service["service_description_url"] = (
            {"nl": service_description_url} if service_description_url else None
        )

    return create_service_catalogus(settings)


def generate_eherkenning_metadata():
    client = eHerkenningClient()
    client.saml2_setting_kwargs = {"sp_validation_only": True}
    metadata = client.create_metadata()
    return (
        b'<?xml version="1.0" encoding="UTF-8"?>\n' + metadata
        if not metadata.startswith(b"<?xml")
        else metadata
    )


def xml_datetime(d):
    return d.isoformat(timespec="seconds")


def create_language_elements(
    element_name: str,
    option_value: Union[dict[str, str], str, None],
    languages: Union[list[str], None] = None,
) -> list[Element]:
    """
    Convert a configuration option into zero or more eHerkenning dienstcatalogus
    elements

    :param element_name: Name of the XML element to be generated
    :param option_value: Configuration option being either a string or a dictionary
                        containing the language code as key, and the option as value.
    :param languages: A node for each language will be made if option_value is a string
    :return list: of etree elements
    """
    default_language = "en"

    if option_value is None:
        return []

    if isinstance(option_value, str):
        option_in_different_langs = (
            {language: option_value for language in languages}
            if languages
            else {default_language: option_value}
        )
    else:
        option_in_different_langs = option_value

    elements = []
    for lang, option in option_in_different_langs.items():
        xml_lang = {"{http://www.w3.org/XML/1998/namespace}lang": lang}
        elements.append(
            ESC(element_name, option, **xml_lang),
        )
    return elements


def create_service_catalogue(id, issue_instant, signature, service_provider):
    ns = namespaces["esc"]
    args = [
        signature,
        service_provider,
    ]
    kwargs = {
        "ID": id,
        f"{{{ns}}}IssueInstant": xml_datetime(issue_instant),
        f"{{{ns}}}Version": "urn:etoegang:1.13:53",
    }
    return ESC("ServiceCatalogue", *args, **kwargs)


def create_signature(id):
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Digital+signature
    """
    transforms = [
        DS(
            "Transform",
            Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature",
        ),
        DS("Transform", Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"),
    ]

    args = [
        DS(
            "SignedInfo",
            DS(
                "CanonicalizationMethod",
                Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#",
            ),
            DS(
                "SignatureMethod",
                Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            ),
            DS(
                "Reference",
                DS("Transforms", *transforms),
                DS("DigestMethod", Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DS("DigestValue"),
                URI=f"#{id}",
            ),
        ),
        DS("SignatureValue"),
    ]
    return DS("Signature", *args)


def create_service_provider(
    service_provider_id: str,
    organization_display_name: str,
    service_definitions: list,
    service_instances: list,
) -> Element:
    ns = namespaces["esc"]
    org_name_elements = create_language_elements(
        "OrganizationDisplayName", organization_display_name
    )

    args = [
        ESC("ServiceProviderID", service_provider_id),
        *org_name_elements,
        *service_definitions,
        *service_instances,
    ]
    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceProvider", *args, **kwargs)


def create_service_definition(
    service_uuid,
    service_name,
    service_description,
    service_description_url,
    loa,
    entity_concerned_types_allowed,
    requested_attributes,
    makelaar_oin,
    service_restrictions_allowed,
):
    service_name_elements = create_language_elements("ServiceName", service_name)
    service_description_elements = create_language_elements(
        "ServiceDescription", service_description
    )
    service_description_url_elements = create_language_elements(
        "ServiceDescriptionURL", service_description_url
    )

    ns = namespaces["esc"]
    args = [
        ESC("ServiceUUID", service_uuid),
        *service_name_elements,
        *service_description_elements,
        *service_description_url_elements,
        SAML("AuthnContextClassRef", loa),
        ESC("HerkenningsmakelaarId", makelaar_oin),
    ]

    for entity in entity_concerned_types_allowed:
        assert isinstance(entity, dict)

        kwargs = {}
        set_number = entity.get("set_number")
        if set_number:
            kwargs["setNumber"] = set_number
        args.append(
            ESC("EntityConcernedTypesAllowed", entity["name"], **kwargs),
        )

    if service_restrictions_allowed:
        args.append(ESC("ServiceRestrictionsAllowed", service_restrictions_allowed))

    for requested_attribute in requested_attributes:
        if isinstance(requested_attribute, dict):
            ra_kwargs = {}
            if "required" in requested_attribute:
                ra_kwargs["isRequired"] = (
                    "true" if requested_attribute["required"] else "false"
                )

            ra_args = []
            if not "purpose_statements" in requested_attribute:
                ra_args += create_language_elements("PurposeStatement", service_name)
            else:
                ra_args += create_language_elements(
                    "PurposeStatement", requested_attribute["purpose_statements"]
                )

            ra_kwargs["Name"] = requested_attribute["name"]
            args.append(
                ESC("RequestedAttribute", *ra_args, **ra_kwargs),
            )
        else:
            args.append(
                ESC("RequestedAttribute", Name=requested_attribute),
            )

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceDefinition", *args, **kwargs)


def create_service_instance(
    service_id,
    service_uuid,
    instance_of_service,
    service_url,
    privacy_policy_url,
    herkenningsmakelaars_id,
    classifiers,
    key_descriptor,
):
    ns = namespaces["esc"]

    service_url_elements = create_language_elements(
        "ServiceURL", service_url, languages=["nl", "en"]
    )
    privacy_url_elements = create_language_elements(
        "PrivacyPolicyURL", privacy_policy_url, languages=["nl", "en"]
    )

    args = [
        ESC("ServiceID", service_id),
        ESC("ServiceUUID", service_uuid),
        ESC("InstanceOfService", instance_of_service),
        *service_url_elements,
        *privacy_url_elements,
        ESC("HerkenningsmakelaarId", herkenningsmakelaars_id),
        ESC("SSOSupport", "false"),
        ESC("ServiceCertificate", key_descriptor),
    ]
    if len(classifiers) > 0:
        args.append(create_classifiers_element(classifiers))

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceInstance", *args, **kwargs)


def create_classifiers_element(classifiers: list) -> ElementMaker:
    classifiers_elements = []
    for classifier in classifiers:
        classifiers_elements.append(ESC("Classifier", classifier))
    return ESC("Classifiers", *classifiers_elements)


def create_key_descriptor(x509_certificate_content: bytes) -> ElementMaker:
    certificate = load_pem_x509_certificate(x509_certificate_content)
    key_name = binascii.hexlify(
        certificate.fingerprint(certificate.signature_hash_algorithm)
    ).decode("ascii")

    # grab the actual base64 data describding the certificate, but without the
    # BEGIN/END CERTIFICATE headers and footers and stripped of line breaks.
    certificate_content = certificate.public_bytes(serialization.Encoding.DER)
    key_descriptor_cert = b64encode(certificate_content).decode("ascii")

    args = [
        DS(
            "KeyInfo",
            DS("KeyName", key_name),
            DS("X509Data", DS("X509Certificate", key_descriptor_cert)),
        )
    ]
    kwargs = {"use": "encryption"}
    return MD("KeyDescriptor", *args, **kwargs)


def create_service_catalogus(conf: EHerkenningConfig, validate: bool = True) -> bytes:
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Service+catalog
    """
    with conf["cert_file"].open("rb") as cert_file:
        x509_certificate_content: bytes = cert_file.read()

    sc_id = str(uuid4())
    service_provider_id = conf["oin"]
    organization_display_name = conf["organization_name"]

    service_definitions = []
    service_instances = []
    for service in conf["services"]:
        key_descriptor = create_key_descriptor(x509_certificate_content)

        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceUUID
        service_uuid = service["service_uuid"]
        service_name = service["service_name"]
        service_description = service["service_description"]
        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceID
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_instance_uuid = service["service_instance_uuid"]

        service_url = service.get(
            "service_url",
        )
        privacy_policy_url = service.get(
            "privacy_policy_url",
        )
        service_description_url = service.get(
            "service_description_url",
        )
        herkenningsmakelaars_id = service.get(
            "herkenningsmakelaars_id",
        )
        entity_concerned_types_allowed = service.get("entity_concerned_types_allowed")
        service_restrictions_allowed = service.get("service_restrictions_allowed")
        requested_attributes = service.get("requested_attributes", [])
        classifiers = service.get("classifiers", [])

        service_definition = create_service_definition(
            service_uuid,
            service_name,
            service_description,
            service_description_url,
            # https://afsprakenstelsel.etoegang.nl/display/as/Level+of+assurance
            service["loa"],
            entity_concerned_types_allowed,
            requested_attributes,
            herkenningsmakelaars_id,
            service_restrictions_allowed,
        )
        service_instance = create_service_instance(
            service_id,
            service_instance_uuid,
            service_uuid,
            service_url,
            privacy_policy_url,
            herkenningsmakelaars_id,
            classifiers,
            key_descriptor,
        )

        service_definitions.append(service_definition)
        service_instances.append(service_instance)

    signature = create_signature(sc_id)
    service_provider = create_service_provider(
        service_provider_id,
        organization_display_name,
        service_definitions,
        service_instances,
    )
    xml = create_service_catalogue(sc_id, timezone.now(), signature, service_provider)

    catalogus = tostring(xml, pretty_print=True, xml_declaration=True, encoding="utf-8")
    if validate:
        errors = validate_xml(BytesIO(catalogus), EHERKENNING_DS_XSD)
        assert errors is None, errors
    return catalogus


def get_metadata_eherkenning_requested_attributes(
    conf: ServiceConfig, service_id: str
) -> list[dict]:
    # There needs to be a RequestedAttribute element where the name is the ServiceID
    # https://afsprakenstelsel.etoegang.nl/display/as/DV+metadata+for+HM
    requested_attributes = [{"name": service_id, "isRequired": False}]
    for requested_attribute in conf.get("requested_attributes", []):
        if isinstance(requested_attribute, dict):
            requested_attributes.append(
                {
                    "name": requested_attribute["name"],
                    "isRequired": requested_attribute["required"],
                }
            )
        else:
            requested_attributes.append(
                {
                    "name": requested_attribute,
                    "isRequired": True,
                }
            )

    return requested_attributes


def create_attribute_consuming_services(conf: EHerkenningConfig) -> list[dict]:
    attribute_consuming_services = []

    for service in conf["services"]:
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_name = get_service_name(service)
        service_description = get_service_description(service)
        requested_attributes = get_metadata_eherkenning_requested_attributes(
            service, service_id
        )

        attribute_consuming_services.append(
            {
                "index": service["attribute_consuming_service_index"],
                "serviceName": service_name,
                "serviceDescription": service_description,
                "requestedAttributes": requested_attributes,
                "language": service.get("language", "nl"),
            }
        )
    return attribute_consuming_services


class eHerkenningClient(BaseSaml2Client):
    cache_key_prefix = "eherkenning"
    cache_timeout = 60 * 60  # 1 hour

    @property
    def conf(self) -> EHerkenningConfig:
        if not hasattr(self, "_conf"):
            db_config = EherkenningConfiguration.get_solo()
            self._conf = db_config.as_dict()
            self._conf.setdefault("acs_path", reverse("eherkenning:acs"))
        return self._conf

    @no_type_check  # my editor has more red than the red wedding in GOT
    def create_config_dict(self, conf: EHerkenningConfig) -> EHerkenningSAMLConfig:
        config_dict: EHerkenningSAMLConfig = super().create_config_dict(conf)

        sp_config = config_dict["sp"]

        # we have multiple services, so delete the config for the "single service" variant
        attribute_consuming_services = create_attribute_consuming_services(conf)
        del sp_config["attributeConsumingService"]
        sp_config["attributeConsumingServices"] = attribute_consuming_services

        return config_dict

    def create_config(
        self, config_dict: EHerkenningSAMLConfig
    ) -> OneLogin_Saml2_Settings:
        config_dict["security"].update(
            {
                # See comment in the python3-saml for in  OneLogin_Saml2_Response.validate_num_assertions (onelogin/saml2/response.py)
                # for why we need this option.
                "disableSignatureWrappingProtection": True,
                # For eHerkenning, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContext": False,
            }
        )
        return super().create_config(config_dict)

    def create_authn_request(
        self, request, return_to=None, attr_consuming_service_index=None, **kwargs
    ):
        return super().create_authn_request(
            request,
            return_to=return_to,
            force_authn=True,
            is_passive=False,
            set_nameid_policy=False,
            name_id_value_req=None,
            attr_consuming_service_index=attr_consuming_service_index,
        )



# Generated by Django 4.2.13 on 2024-07-19 12:34

from django.db import migrations

from ..choices import ConfigTypes


def move_certificates(apps, _):
    DigidConfiguration = apps.get_model("digid_eherkenning", "DigidConfiguration")
    EherkenningConfiguration = apps.get_model(
        "digid_eherkenning", "EherkenningConfiguration"
    )
    ConfigCertificate = apps.get_model("digid_eherkenning", "ConfigCertificate")

    for model in (DigidConfiguration, EherkenningConfiguration):
        config = model.objects.first()
        if config is None or not (cert := config.certificate):
            continue
        ConfigCertificate.objects.get_or_create(
            certificate=cert,
            config_type=ConfigTypes(
                f"{config._meta.app_label}.{config._meta.object_name}"
            ),
        )


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning",
            "0011_configcertificate_configcertificate_uniq_config_cert",
        ),
    ]

    operations = [
        # reverse migration is ambiguous, if needed, you can easily use the UI
        migrations.RunPython(move_certificates, migrations.RunPython.noop),
        migrations.RemoveField(
            model_name="digidconfiguration",
            name="certificate",
        ),
        migrations.RemoveField(
            model_name="eherkenningconfiguration",
            name="certificate",
        ),
    ]



from .certificates import ConfigCertificate
from .digid import DigidConfiguration
from .eherkenning import EherkenningConfiguration

__all__ = [
    "DigidConfiguration",
    "EherkenningConfiguration",
    "ConfigCertificate",
]



from django.core.management import BaseCommand

from ...models import DigidConfiguration, EherkenningConfiguration

MODEL_MAP = {
    "digid": DigidConfiguration,
    "eherkenning": EherkenningConfiguration,
}


class Command(BaseCommand):
    help = "Updates the stored metadata file and prepopulates the db fields."

    def add_arguments(self, parser):
        parser.add_argument(
            "config_model",
            type=str,
            choices=list(MODEL_MAP.keys()),
            help="Update the DigiD or Eherkenning configuration metadata.",
        )

    def handle(self, **options):
        config_model = MODEL_MAP[options["config_model"]]
        config = config_model.get_solo()

        if config.metadata_file_source:
            config.save(force_metadata_update=True)
            self.stdout.write(self.style.SUCCESS("Update was successful"))
        else:
            self.stdout.write(
                self.style.WARNING("Update failed, no metadata file source found")
            )



# Generated by Django 4.2.13 on 2024-06-12 06:50

from django.db import migrations, models

import django_jsonform.models.fields

import digid_eherkenning.oidc.models.base


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning_oidc_generics",
            "0005_digidmachtigenconfig_mandate_service_id_claim_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="digidconfig",
            name="oidc_rp_scopes_list",
            field=django_jsonform.models.fields.ArrayField(
                base_field=models.CharField(
                    max_length=50, verbose_name="OpenID Connect scope"
                ),
                blank=True,
                default=digid_eherkenning.oidc.models.base.get_default_scopes_bsn,
                help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider.",
                size=None,
                verbose_name="OpenID Connect scopes",
            ),
        ),
        migrations.AlterField(
            model_name="digidmachtigenconfig",
            name="oidc_rp_scopes_list",
            field=django_jsonform.models.fields.ArrayField(
                base_field=models.CharField(
                    max_length=50, verbose_name="OpenID Connect scope"
                ),
                blank=True,
                default=digid_eherkenning.oidc.models.base.get_default_scopes_bsn,
                help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider.",
                size=None,
                verbose_name="OpenID Connect scopes",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningbewindvoeringconfig",
            name="oidc_rp_scopes_list",
            field=django_jsonform.models.fields.ArrayField(
                base_field=models.CharField(
                    max_length=50, verbose_name="OpenID Connect scope"
                ),
                blank=True,
                default=digid_eherkenning.oidc.models.base.get_default_scopes_bsn,
                help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider.",
                size=None,
                verbose_name="OpenID Connect scopes",
            ),
        ),
        migrations.AlterField(
            model_name="eherkenningconfig",
            name="oidc_rp_scopes_list",
            field=django_jsonform.models.fields.ArrayField(
                base_field=models.CharField(
                    max_length=50, verbose_name="OpenID Connect scope"
                ),
                blank=True,
                default=digid_eherkenning.oidc.models.base.get_default_scopes_kvk,
                help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider.",
                size=None,
                verbose_name="OpenID Connect scopes",
            ),
        ),
    ]



import logging

from glom import Path, PathAccessError, glom
from mozilla_django_oidc_db.typing import ClaimPath, JSONObject

from .models import BaseConfig

logger = logging.getLogger(__name__)


class NoLOAClaim(Exception):
    pass


def process_claims(
    claims: JSONObject,
    config: BaseConfig,
    strict: bool = True,
) -> JSONObject:
    """
    Given the raw claims, process them using the provided config.

    Claim processing performs the following steps:

    * Claim name normalization, the provided config model field names are used as keys
    * Extracting required and optional values. An error is thrown for missing required
      claims, unless a default value is specified in the config.
    * Claim value post-processing - if values need to be translated/normalized, the
      provided configuration is used.

    The return value SHOULD include the ``loa_claim`` key, but if no value is available
    (not in the claims and no default specified -> then it's omitted), the key will be
    absent.

    :arg claims: The raw claims as received from the Identity Provider.
    :arg config: The OIDC Configuration instance that specifies which claims should be
      extracted and processed.
    :arg strict: In strict mode, absent claims that are required (according) to the
      configuration raise an error. In non-strict mode, these claims are simply skipped
      and omitted.
    :returns: A (JSON-serializable) dictionary where the keys are the claim config
      field names, taken from ``config.CLAIMS_CONFIGURATION``, and the values their
      extracted values from the raw claims. Extracted values have been post-processed
      if post-processing configuration was available.
    """
    processed_claims = {}

    # first, extract all the configured required claims
    for claim_config in config.CLAIMS_CONFIGURATION:
        field_name = claim_config["field"]
        path_bits: ClaimPath = getattr(config, field_name)
        try:
            value = glom(claims, Path(*path_bits))
        except PathAccessError as exc:
            if not claim_config["required"]:
                continue
            # in non-strict mode, do not raise but instead omit the claim. Up to the
            # caller to handle missing claims.
            if not strict:
                continue
            claim_repr = " > ".join(path_bits)
            raise ValueError(f"Required claim '{claim_repr}' not found") from exc

        processed_claims[field_name] = value

    # then, loa is hardcoded in the base model, process those...
    try:
        loa = _process_loa(claims, config)
    except NoLOAClaim as exc:
        logger.info(
            "Missing LoA claim, excluding it from processed claims", exc_info=exc
        )
    else:
        processed_claims["loa_claim"] = loa

    return processed_claims


def _process_loa(claims: JSONObject, config: BaseConfig) -> str:
    default = config.default_loa
    if not (loa_claim := config.loa_claim) and not default:
        raise NoLOAClaim("No LoA claim or default LoA configured")

    if not loa_claim:
        return default

    try:
        loa = glom(claims, Path(*config.loa_claim))
        loa_claim_missing = False
    except PathAccessError:
        # default could be empty (string)!
        loa = default
        loa_claim_missing = not default

    if loa_claim_missing:
        raise NoLOAClaim("LoA claim is absent and no default LoA configured")

    # 'from' is string or number, which are valid keys
    loa_map: dict[str | float | int, str] = {
        mapping["from"]: mapping["to"] for mapping in config.loa_value_mapping
    }

    # apply mapping, if not found -> use the literal original value instead
    return loa_map.get(loa, loa)



# Generated by Django 3.2.20 on 2023-09-08 10:43

from django.db import migrations, models

import django_jsonform.models.fields
import mozilla_django_oidc_db.models

import digid_eherkenning.oidc.models


class Migration(migrations.Migration):
    dependencies = []

    operations = [
        migrations.CreateModel(
            name="OpenIDConnectDigiDMachtigenConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_op_logout_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider logout endpoint",
                        max_length=1000,
                        verbose_name="Logout endpoint",
                    ),
                ),
                (
                    "oidc_keycloak_idp_hint",
                    models.CharField(
                        blank=True,
                        help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                        max_length=1000,
                        verbose_name="Keycloak Identity Provider hint",
                    ),
                ),
                (
                    "vertegenwoordigde_claim_name",
                    models.CharField(
                        default="aanvrager.bsn",
                        help_text="Name of the claim in which the BSN of the person being represented is stored",
                        max_length=50,
                        verbose_name="vertegenwoordigde claim name",
                    ),
                ),
                (
                    "gemachtigde_claim_name",
                    models.CharField(
                        default="gemachtigde.bsn",
                        help_text="Name of the claim in which the BSN of the person representing someone else is stored",
                        max_length=50,
                        verbose_name="gemachtigde claim name",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=digid_eherkenning.oidc.models.get_default_scopes_bsn,
                        help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "oidc_exempt_urls",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=1000, verbose_name="Exempt URL"
                        ),
                        blank=True,
                        default=list,
                        help_text="This is a list of absolute url paths, regular expressions for url paths, or Django view names. This plus the mozilla-django-oidc urls are exempted from the session renewal by the SessionRefresh middleware.",
                        size=None,
                        verbose_name="URLs exempt from session renewal",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect state verification",
                        verbose_name="State size",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        help_text="Controls whether the OpenID Connect client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration for DigiD Machtigen",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="OpenIDConnectEHerkenningBewindvoeringConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        help_text="Controls whether the OpenID Connect client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect state verification",
                        verbose_name="State size",
                    ),
                ),
                (
                    "oidc_exempt_urls",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=1000, verbose_name="Exempt URL"
                        ),
                        blank=True,
                        default=list,
                        help_text="This is a list of absolute url paths, regular expressions for url paths, or Django view names. This plus the mozilla-django-oidc urls are exempted from the session renewal by the SessionRefresh middleware.",
                        size=None,
                        verbose_name="URLs exempt from session renewal",
                    ),
                ),
                (
                    "oidc_op_logout_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider logout endpoint",
                        max_length=1000,
                        verbose_name="Logout endpoint",
                    ),
                ),
                (
                    "oidc_keycloak_idp_hint",
                    models.CharField(
                        blank=True,
                        help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                        max_length=1000,
                        verbose_name="Keycloak Identity Provider hint",
                    ),
                ),
                (
                    "vertegenwoordigde_company_claim_name",
                    models.CharField(
                        default="aanvrager.kvk",
                        help_text="Name of the claim in which the KVK of the company being represented is stored",
                        max_length=50,
                        verbose_name="vertegenwoordigde company claim name",
                    ),
                ),
                (
                    "gemachtigde_person_claim_name",
                    models.CharField(
                        default="gemachtigde.pseudoID",
                        help_text="Name of the claim in which the ID of the person representing a company is stored",
                        max_length=50,
                        verbose_name="gemachtigde person claim name",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=digid_eherkenning.oidc.models.get_default_scopes_bsn,
                        help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration for eHerkenning Bewindvoering",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="OpenIDConnectEHerkenningConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_op_logout_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider logout endpoint",
                        max_length=1000,
                        verbose_name="Logout endpoint",
                    ),
                ),
                (
                    "oidc_keycloak_idp_hint",
                    models.CharField(
                        blank=True,
                        help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                        max_length=1000,
                        verbose_name="Keycloak Identity Provider hint",
                    ),
                ),
                (
                    "identifier_claim_name",
                    models.CharField(
                        default="kvk",
                        help_text="The name of the claim in which the KVK of the user is stored",
                        max_length=100,
                        verbose_name="KVK claim name",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=digid_eherkenning.oidc.models.get_default_scopes_kvk,
                        help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "oidc_exempt_urls",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=1000, verbose_name="Exempt URL"
                        ),
                        blank=True,
                        default=list,
                        help_text="This is a list of absolute url paths, regular expressions for url paths, or Django view names. This plus the mozilla-django-oidc urls are exempted from the session renewal by the SessionRefresh middleware.",
                        size=None,
                        verbose_name="URLs exempt from session renewal",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect state verification",
                        verbose_name="State size",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        help_text="Controls whether the OpenID Connect client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration for eHerkenning",
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="OpenIDConnectPublicConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_op_discovery_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                        max_length=1000,
                        verbose_name="Discovery endpoint",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm.",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format.",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
                (
                    "oidc_op_logout_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider logout endpoint",
                        max_length=1000,
                        verbose_name="Logout endpoint",
                    ),
                ),
                (
                    "oidc_keycloak_idp_hint",
                    models.CharField(
                        blank=True,
                        help_text="Specific for Keycloak: parameter that indicates which identity provider should be used (therefore skipping the Keycloak login screen).",
                        max_length=1000,
                        verbose_name="Keycloak Identity Provider hint",
                    ),
                ),
                (
                    "identifier_claim_name",
                    models.CharField(
                        default="bsn",
                        help_text="The name of the claim in which the BSN of the user is stored",
                        max_length=100,
                        verbose_name="BSN claim name",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=digid_eherkenning.oidc.models.get_default_scopes_bsn,
                        help_text="OpenID Connect scopes that are requested during login. These scopes are hardcoded and must be supported by the identity provider",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "oidc_exempt_urls",
                    django_jsonform.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=1000, verbose_name="Exempt URL"
                        ),
                        blank=True,
                        default=list,
                        help_text="This is a list of absolute url paths, regular expressions for url paths, or Django view names. This plus the mozilla-django-oidc urls are exempted from the session renewal by the SessionRefresh middleware.",
                        size=None,
                        verbose_name="URLs exempt from session renewal",
                    ),
                ),
                (
                    "oidc_nonce_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect nonce verification",
                        verbose_name="Nonce size",
                    ),
                ),
                (
                    "oidc_state_size",
                    models.PositiveIntegerField(
                        default=32,
                        help_text="Sets the length of the random string used for OpenID Connect state verification",
                        verbose_name="State size",
                    ),
                ),
                (
                    "oidc_use_nonce",
                    models.BooleanField(
                        default=True,
                        help_text="Controls whether the OpenID Connect client uses nonce verification",
                        verbose_name="Use nonce",
                    ),
                ),
                (
                    "userinfo_claims_source",
                    models.CharField(
                        choices=[
                            ("userinfo_endpoint", "Userinfo endpoint"),
                            ("id_token", "ID token"),
                        ],
                        default="userinfo_endpoint",
                        help_text="Indicates the source from which the user information claims should be extracted.",
                        max_length=100,
                        verbose_name="user information claims extracted from",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration for DigiD",
            },
            bases=(models.Model,),
        ),
    ]



from datetime import datetime

from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from privates.admin import PrivateMediaMixin
from privates.widgets import PrivateFileWidget
from solo.admin import SingletonModelAdmin

from .models import ConfigCertificate, DigidConfiguration, EherkenningConfiguration
from .models.base import BaseConfiguration


class CustomPrivateFileWidget(PrivateFileWidget):
    template_name = "admin/digid_eherkenning/widgets/custom_file_input.html"


class CustomPrivateMediaMixin(PrivateMediaMixin):
    private_media_file_widget = CustomPrivateFileWidget


class BaseAdmin(CustomPrivateMediaMixin, SingletonModelAdmin):
    readonly_fields = (
        "link_to_certificates",
        "idp_service_entity_id",
    )
    private_media_fields = ("idp_metadata_file",)

    @admin.display(description=_("certificates"))
    def link_to_certificates(self, obj: BaseConfiguration) -> str:
        path = reverse(
            "admin:digid_eherkenning_configcertificate_changelist",
            current_app=self.admin_site.name,
        )
        config_type = obj._as_config_type()
        qs = ConfigCertificate.objects.filter(config_type=config_type)
        url = f"{path}?config_type__exact={config_type}"
        return format_html(
            '<a href="{url}" target="_blank">{label}</a>',
            url=url,
            config_type=config_type.value,
            label=_("Manage ({count})").format(count=qs.count()),
        )


def _fieldset_factory(middle):
    """
    Output custom fieldsets (model-specific) between fixed shared field(set)s.
    """
    head = [
        (
            _("X.509 Certificate"),
            {
                "fields": ("link_to_certificates",),
            },
        ),
        (
            _("Identity provider"),
            {
                "fields": (
                    "metadata_file_source",
                    "idp_service_entity_id",
                    "idp_metadata_file",
                ),
            },
        ),
        (
            _("SAML configuration"),
            {
                "fields": (
                    "entity_id",
                    "base_url",
                    "artifact_resolve_content_type",
                    "want_assertions_signed",
                    "want_assertions_encrypted",
                    "signature_algorithm",
                    "digest_algorithm",
                ),
            },
        ),
    ]
    tail = [
        (
            _("Organization details"),
            {
                "fields": (
                    "technical_contact_person_telephone",
                    "technical_contact_person_email",
                    "organization_url",
                    "organization_name",
                ),
            },
        ),
    ]

    return tuple(head + list(middle) + tail)


@admin.register(DigidConfiguration)
class DigidConfigurationAdmin(BaseAdmin):
    fieldsets = _fieldset_factory(
        [
            (
                _("Service details"),
                {
                    "fields": (
                        "service_name",
                        "service_description",
                        "requested_attributes",
                        "attribute_consuming_service_index",
                        "slo",
                    ),
                },
            ),
        ]
    )
    change_form_template = "admin/digid_eherkenning/digidconfiguration/change_form.html"


@admin.register(EherkenningConfiguration)
class EherkenningConfigurationAdmin(BaseAdmin):
    fieldsets = _fieldset_factory(
        [
            (
                _("Service details"),
                {
                    "fields": (
                        "service_name",
                        "service_description",
                        "oin",
                        "makelaar_id",
                        "privacy_policy",
                        "service_language",
                    ),
                },
            ),
            (
                _("eHerkenning"),
                {
                    "fields": (
                        "eh_requested_attributes",
                        "eh_attribute_consuming_service_index",
                        "eh_service_uuid",
                        "eh_service_instance_uuid",
                        "eh_loa",
                    ),
                },
            ),
            (
                _("eIDAS"),
                {
                    "fields": (
                        "no_eidas",
                        "eidas_requested_attributes",
                        "eidas_attribute_consuming_service_index",
                        "eidas_service_uuid",
                        "eidas_service_instance_uuid",
                        "eidas_loa",
                    ),
                },
            ),
        ]
    )

    change_form_template = (
        "admin/digid_eherkenning/eherkenningconfiguration/change_form.html"
    )


@admin.register(ConfigCertificate)
class ConfigCertificateAdmin(admin.ModelAdmin):
    list_display = (
        "config_type",
        "certificate",
        "valid_from",
        "expiry_date",
        "is_ready",
    )
    list_filter = ("config_type",)
    search_fields = ("certificate__label",)
    raw_id_fields = ("certificate",)

    @admin.display(description=_("valid from"))
    def valid_from(self, obj: ConfigCertificate) -> datetime:
        return obj.certificate.valid_from

    @admin.display(description=_("expires on"))
    def expiry_date(self, obj: ConfigCertificate) -> datetime:
        return obj.certificate.expiry_date

    @admin.display(description=_("valid candidate?"), boolean=True)
    def is_ready(self, obj: ConfigCertificate) -> bool:
        return obj.is_ready_for_authn_requests



from collections.abc import Sequence
from copy import deepcopy

from django.contrib import admin
from django.forms import modelform_factory
from django.utils.translation import gettext_lazy as _

from mozilla_django_oidc_db.forms import OpenIDConnectConfigForm
from solo.admin import SingletonModelAdmin

from .models import (
    BaseConfig,
    DigiDConfig,
    DigiDMachtigenConfig,
    EHerkenningBewindvoeringConfig,
    EHerkenningConfig,
)

# Using a dict because these retain ordering, and it makes things a bit more readable.
ATTRIBUTES_MAPPING_TITLE = _("Attributes to extract from claims")
COMMON_FIELDSETS = {
    _("Activation"): {
        "fields": ("enabled",),
    },
    _("Common settings"): {
        "fields": (
            "oidc_rp_client_id",
            "oidc_rp_client_secret",
            "oidc_rp_scopes_list",
            "oidc_rp_sign_algo",
            "oidc_rp_idp_sign_key",
        ),
    },
    ATTRIBUTES_MAPPING_TITLE: {
        "fields": (),  # populated from the factory function below
    },
    _("Endpoints"): {
        "fields": (
            "oidc_op_discovery_endpoint",
            "oidc_op_jwks_endpoint",
            "oidc_op_authorization_endpoint",
            "oidc_op_token_endpoint",
            "oidc_token_use_basic_auth",
            "oidc_op_user_endpoint",
            "oidc_op_logout_endpoint",
        ),
    },
    _("Keycloak specific settings"): {
        "fields": ("oidc_keycloak_idp_hint",),
        "classes": ["collapse in"],
    },
    _("Advanced settings"): {
        "fields": (
            "oidc_use_nonce",
            "oidc_nonce_size",
            "oidc_state_size",
            "userinfo_claims_source",
        ),
        "classes": ["collapse in"],
    },
}


def admin_modelform_factory(model: type[BaseConfig], *args, **kwargs):
    """
    Factory function to generate a model form class for a given configuration model.

    The configuration model is expected to be a subclass of
    :class:`~digid_eherkenning_oidc_generics.models.OpenIDConnectBaseConfig`.

    Additional args and kwargs are forwarded to django's
    :func:`django.forms.modelform_factory`.
    """
    kwargs.setdefault("form", OpenIDConnectConfigForm)
    Form = modelform_factory(model, *args, **kwargs)
    assert issubclass(
        Form, OpenIDConnectConfigForm
    ), "The base form class must be a subclass of OpenIDConnectConfigForm."
    return Form


def fieldsets_factory(claim_mapping_fields: Sequence[str | Sequence[str]]):
    """
    Apply the shared fieldsets configuration with the model-specific overrides.
    """
    _fieldsets = deepcopy(COMMON_FIELDSETS)
    _fieldsets[ATTRIBUTES_MAPPING_TITLE]["fields"] += tuple(claim_mapping_fields)
    return tuple((label, config) for label, config in _fieldsets.items())


@admin.register(DigiDConfig)
class DigiDConfigAdmin(SingletonModelAdmin):
    form = admin_modelform_factory(DigiDConfig)
    fieldsets = fieldsets_factory(
        claim_mapping_fields=[
            "bsn_claim",
            "loa_claim",
            "default_loa",
            "loa_value_mapping",
        ]
    )


@admin.register(EHerkenningConfig)
class EHerkenningConfigAdmin(SingletonModelAdmin):
    form = admin_modelform_factory(EHerkenningConfig)
    fieldsets = fieldsets_factory(
        claim_mapping_fields=[
            "identifier_type_claim",
            "legal_subject_claim",
            "branch_number_claim",
            "acting_subject_claim",
            "loa_claim",
            "default_loa",
            "loa_value_mapping",
        ]
    )


@admin.register(DigiDMachtigenConfig)
class DigiDMachtigenConfigAdmin(SingletonModelAdmin):
    form = admin_modelform_factory(DigiDMachtigenConfig)
    fieldsets = fieldsets_factory(
        claim_mapping_fields=[
            "representee_bsn_claim",
            "authorizee_bsn_claim",
            "loa_claim",
            "default_loa",
            "loa_value_mapping",
            "mandate_service_id_claim",
        ]
    )


@admin.register(EHerkenningBewindvoeringConfig)
class EHerkenningBewindvoeringConfigAdmin(SingletonModelAdmin):
    form = admin_modelform_factory(EHerkenningBewindvoeringConfig)
    fieldsets = fieldsets_factory(
        claim_mapping_fields=[
            "representee_claim",
            "identifier_type_claim",
            "legal_subject_claim",
            "branch_number_claim",
            "acting_subject_claim",
            "loa_claim",
            "default_loa",
            "loa_value_mapping",
            "mandate_service_id_claim",
            "mandate_service_uuid_claim",
        ]
    )



from datetime import datetime, timedelta
from io import BytesIO

from django.core.files import File
from django.utils import timezone

import pytest
from cryptography import x509
from freezegun import freeze_time
from simple_certmanager.constants import CertificateTypes
from simple_certmanager.models import Certificate
from simple_certmanager.test.certificate_generation import (
    cert_to_pem,
    gen_key,
    key_to_pem,
    mkcert,
)

from digid_eherkenning.choices import ConfigTypes
from digid_eherkenning.exceptions import CertificateProblem
from digid_eherkenning.models import ConfigCertificate
from digid_eherkenning.models.digid import DigidConfiguration
from digid_eherkenning.models.eherkenning import EherkenningConfiguration

pytestmark = [pytest.mark.django_db]


def test_valid_certificate(temp_private_root, digid_certificate):
    # note that this test will start failing once the certificates expire IRL (!)
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    assert config_certificate.is_ready_for_authn_requests


@freeze_time("2099-01-01")
def test_expired_certificate(temp_private_root, digid_certificate):
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    assert not config_certificate.is_ready_for_authn_requests


@freeze_time("1700-01-01")
def test_certificate_not_valid_yet(temp_private_root, digid_certificate):
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    assert not config_certificate.is_ready_for_authn_requests


def test_certificate_file_missing():
    # can happen if the infrastructure has an oopsie...
    certificate = Certificate.objects.create(
        type=CertificateTypes.key_pair,
        public_certificate="bad/filepath/cert.pem",
    )
    assert certificate.public_certificate.path
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=certificate,
    )

    assert not config_certificate.is_ready_for_authn_requests


def test_instance_without_certificate_provided():
    config_certificate = ConfigCertificate(config_type=ConfigTypes.digid)

    assert not config_certificate.is_ready_for_authn_requests


def test_certificate_wrong_type(temp_private_root, digid_certificate):
    digid_certificate.type = CertificateTypes.cert_only
    digid_certificate.save()
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    assert not config_certificate.is_ready_for_authn_requests


@pytest.mark.parametrize("path", ("", "missing/dir/bad-key-path.pem"))
def test_private_key_missing(temp_private_root, digid_certificate, path):
    digid_certificate.private_key = path
    digid_certificate.save()
    config_certificate = ConfigCertificate(
        config_type=ConfigTypes.digid,
        certificate=digid_certificate,
    )

    assert not config_certificate.is_ready_for_authn_requests


def test_string_representation(settings, digid_certificate):
    settings.LANGUAGE_CODE = "en"
    digid_certificate.label = "SAML"
    cc1 = ConfigCertificate(
        config_type=ConfigTypes.digid, certificate=digid_certificate
    )
    assert str(cc1) == "DigiD: SAML"

    cc2 = ConfigCertificate(config_type=ConfigTypes.eherkenning, certificate=None)
    assert str(cc2) == "eHerkenning: (no certificate selected)"


# Helpers for multiple certificates - can't call fixtures multiple times to get
# different outcomes.


def _generate_config_certificate(
    request: pytest.FixtureRequest,
    config_type: ConfigTypes,
    valid_from: datetime,
) -> Certificate:
    request.getfixturevalue("temp_private_root")
    subject = x509.Name(
        [
            x509.NameAttribute(x509.NameOID.COUNTRY_NAME, "NL"),
            x509.NameAttribute(x509.NameOID.STATE_OR_PROVINCE_NAME, "Some-State"),
            x509.NameAttribute(x509.NameOID.ORGANIZATION_NAME, "OpenGem"),
            x509.NameAttribute(x509.NameOID.COMMON_NAME, "widgits.example.org"),
        ]
    )
    root_cert = request.getfixturevalue("root_cert")
    root_key = request.getfixturevalue("root_key")

    key = gen_key()
    # hack to work around the hardcoded timezone.now() usage in the mkcert helper
    assert valid_from.tzinfo is not None, "Thou shall not use naive datetimes"
    with freeze_time(valid_from):
        cert = mkcert(
            subject=subject,
            subject_key=key,
            issuer=root_cert,
            issuer_key=root_key,
            can_issue=False,
        )

    cert_pem, key_pem = cert_to_pem(cert), key_to_pem(key)
    certificate = Certificate.objects.create(
        label=f"Certificate, {valid_from.isoformat()}",
        type=CertificateTypes.key_pair,
        public_certificate=File(BytesIO(cert_pem), name="public_certificate.pem"),
        private_key=File(BytesIO(key_pem), name="private_key.pem"),
    )
    ConfigCertificate.objects.create(config_type=config_type, certificate=certificate)
    return certificate


@pytest.mark.parametrize("model", (DigidConfiguration, EherkenningConfiguration))
def test_certificate_selection_picks_correct(
    request: pytest.FixtureRequest,
    model: type[DigidConfiguration] | type[EherkenningConfiguration],
):
    config = model.get_solo()
    config_type = config._as_config_type()
    # expired
    _generate_config_certificate(
        request, config_type, valid_from=timezone.now() - timedelta(days=5)
    )
    # currently valid
    _current_cert = _generate_config_certificate(
        request, config_type, valid_from=timezone.now()
    )
    # valid tomorrow
    _next_cert = _generate_config_certificate(
        request,
        config_type,
        valid_from=timezone.now() + timedelta(days=1),
    )

    current_cert, next_cert = config.select_certificates()

    assert current_cert == _current_cert
    assert next_cert == _next_cert


@pytest.mark.parametrize("model", (DigidConfiguration, EherkenningConfiguration))
def test_certificate_selection_picks_correct_2(
    request: pytest.FixtureRequest,
    model: type[DigidConfiguration] | type[EherkenningConfiguration],
):
    config = model.get_solo()
    config_type = config._as_config_type()
    # expired
    _generate_config_certificate(
        request, config_type, valid_from=timezone.now() - timedelta(days=5)
    )
    # currently valid
    _old_current = _generate_config_certificate(
        request, config_type, valid_from=timezone.now()
    )
    # valid tomorrow
    _new_current = _generate_config_certificate(
        request,
        config_type,
        valid_from=timezone.now() + timedelta(days=1),
    )

    # "current" has now expired
    with freeze_time(timezone.now() + timedelta(days=1, hours=1)):
        assert timezone.now() > _old_current.expiry_date
        current_cert, next_cert = config.select_certificates()

    assert current_cert == _new_current
    assert next_cert is None


@pytest.mark.parametrize("model", (DigidConfiguration, EherkenningConfiguration))
def test_no_current_certificate(
    request: pytest.FixtureRequest,
    model: type[DigidConfiguration] | type[EherkenningConfiguration],
):
    config = model.get_solo()
    config_type = config._as_config_type()
    # expired
    _generate_config_certificate(
        request, config_type, valid_from=timezone.now() - timedelta(days=5)
    )

    with pytest.raises(CertificateProblem):
        config.select_certificates()


@pytest.mark.parametrize("model", (DigidConfiguration, EherkenningConfiguration))
def test_skips_invalid_certificates_for_current(
    request: pytest.FixtureRequest,
    model: type[DigidConfiguration] | type[EherkenningConfiguration],
):
    config = model.get_solo()
    config_type = config._as_config_type()
    now = timezone.now()

    # all are currently valid - but we'll introduce problems
    (c1, c2, c3, c4, c5, c6) = [
        _generate_config_certificate(request, config_type, valid_from=now)
        for _ in range(0, 6)
    ]
    # must be keypair
    c1.type = CertificateTypes.cert_only
    c1.save()
    c2.private_key = ""
    c2.save()
    c3.public_certificate = ""
    c3.save()
    c4.public_certificate.storage.delete(c4.public_certificate.name)
    c5.private_key.storage.delete(
        c5.public_certificate.name
    )  # introduce mismatch between key and certificate
    c6.public_certificate = c1.public_certificate
    c6.save()

    # this one is okay
    _current = _generate_config_certificate(request, config_type, valid_from=now)

    current_cert, next_cert = config.select_certificates()

    assert current_cert == _current
    assert next_cert is None


@pytest.mark.parametrize("model", (DigidConfiguration, EherkenningConfiguration))
def test_skips_invalid_certificates_for_next(
    request: pytest.FixtureRequest,
    model: type[DigidConfiguration] | type[EherkenningConfiguration],
):
    config = model.get_solo()
    config_type = config._as_config_type()
    now = timezone.now()

    # all are currently valid - but we'll introduce problems
    (c1, c2, c3, c4, c5) = [
        _generate_config_certificate(
            request, config_type, valid_from=now + timedelta(days=1)
        )
        for _ in range(0, 5)
    ]
    # must be keypair
    c1.type = CertificateTypes.cert_only
    c1.save()
    c2.private_key = ""
    c2.save()
    c3.public_certificate = ""
    c3.save()
    c4.public_certificate.storage.delete(c4.public_certificate.name)
    c5.private_key.storage.delete(c5.public_certificate.name)
    # this one is okay
    _current = _generate_config_certificate(request, config_type, valid_from=now)

    current_cert, next_cert = config.select_certificates()

    assert current_cert == _current
    assert next_cert is None



"""
DigiD-eHerkenning-OIDC-generics abstracts authenticating over OIDC.

DigiD/eHerkenning are typically exposed via SAML bindings, but there exist identity
providers that abstract this and instead offer an OpenID Connect flow to log in with
DigiD and/or eHerkenning. This package facilitates integrating with such providers.

The architecture and authentication flows are tricky in some places. Here's an attempt
to explain it.

**Configuration**

Each authentication means (DigiD, eHerkenning, mandate (machtigen) variants...) is
mapped to an OpenID client configuration, which roughly holds:

- the OIDC endpoints to use/redirect users to
- the OAUTH2 client ID and secret to use, which indicate to the IdP which authentication
  means they should send the user to
- which claims to look up/extract from the UserInfo endpoint/JWT

These are stored in (subclasses of) the
:class:`~digid_herkenning.oidc.models.OpenIDConnectBaseConfig` model.

**Authentication flow**

When a user starts a login flow, they:

1. Click the appriopriate button/link
2. A Django view processes this and looks up the relevant configuration
3. The view redirects the user to the identity provider (typically a different domain)
4. Authenticate with the IdP
5. The IdP redirects back to our application
6. Our callback view performs the OIDC exchange and extracts + stores the relevant user
   information
7. Finally, the callback view looks up where the user needs to be redirected to and
   sends them that way.

Steps 2-3 are called the "init" phase in this package, while steps 6-7 are the
"callback" phase.

**Init phase**

The mozilla-django-oidc-db package provides the
:class:`~mozilla_django_oidc_db.views.OIDCInit` view class, for the init phase. It
ensures that the specified config class is persisted in the authentication state.

This package provides concrete views bound to configuration classes:

* :attr:`~digid_herkenning.oidc.views.digid_init`
* :attr:`~digid_herkenning.oidc.views.digid_machtigen_init`
* :attr:`~digid_herkenning.oidc.views.eherkenning_init`
* :attr:`~digid_herkenning.oidc.views.eherkenning_bewindvoering_init`

**Callback phase**

The callback phase validates the code and state, and loads which configuration class
needs to be used from the state. With this information, the authentication backends
from ``settings.AUTHENTICATION_BACKENDS`` are tried in order. Typically this will
use the backend shipped in mozilla-django-oidc-db, or a subclass of it.

The OpenID connect flow exchanges the code for an access token (and ID token), and
the user details are retrieved. You should provide a customized backend to determine
what needs to be done with this user information, e.g. create a django user or store
the information in the django session.
"""



# Generated by Django 4.2.13 on 2024-06-11 20:44

from django.db import migrations, models

import mozilla_django_oidc_db.fields


class Migration(migrations.Migration):

    dependencies = [
        (
            "digid_eherkenning_oidc_generics",
            "0004_migrate_config_to_claimfield",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="digidmachtigenconfig",
            name="mandate_service_id_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:nl-eid-gdi:1.0:ServiceUUID"
                ),
                help_text="Name of the claim holding the service UUID for which the acting subject is authorized.",
                size=None,
                verbose_name="service ID claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="branch_number_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:1.9:ServiceRestriction:Vestigingsnr"
                ),
                help_text="Name of the claim holding the value of the branch number for the authenticated company, if such a restriction applies.",
                size=None,
                verbose_name="branch number claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="mandate_service_id_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:ServiceID"
                ),
                help_text="Name of the claim holding the service ID for which the company is authorized.",
                size=None,
                verbose_name="service ID claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningbewindvoeringconfig",
            name="mandate_service_uuid_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:core:ServiceUUID"
                ),
                help_text="Name of the claim holding the service UUID for which the company is authorized.",
                size=None,
                verbose_name="service UUID claim",
            ),
        ),
        migrations.AddField(
            model_name="eherkenningconfig",
            name="branch_number_claim",
            field=mozilla_django_oidc_db.fields.ClaimField(
                base_field=models.CharField(
                    max_length=50, verbose_name="claim path segment"
                ),
                default=mozilla_django_oidc_db.fields.ClaimFieldDefault(
                    "urn:etoegang:1.9:ServiceRestriction:Vestigingsnr"
                ),
                help_text="Name of the claim holding the value of the branch number for the authenticated company, if such a restriction applies.",
                size=None,
                verbose_name="branch number claim",
            ),
        ),
    ]



from django.test import TestCase

import pytest
from lxml import etree
from simple_certmanager.models import Certificate

from digid_eherkenning.choices import ConfigTypes
from digid_eherkenning.models import ConfigCertificate, DigidConfiguration
from digid_eherkenning.saml2.digid import generate_digid_metadata

from .mixins import DigidMetadataMixin

NAME_SPACES = {
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
}


@pytest.mark.usefixtures("digid_config", "temp_private_root")
class DigidMetadataGenerationTests(DigidMetadataMixin, TestCase):
    def test_generate_metadata_all_options_specified(self):
        self.digid_config.want_assertions_signed = True
        self.digid_config.signature_algorithm = (
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256"
        )
        self.digid_config.digest_algorithm = "http://www.w3.org/2001/04/xmlenc#sha256"
        self.digid_config.attribute_consuming_service_index = "9050"
        self.digid_config.technical_contact_person_telephone = "06123123123"
        self.digid_config.technical_contact_person_email = "test@test.nl"
        self.digid_config.organization_name = "Test organisation"
        self.digid_config.organization_url = "http://test-organisation.nl"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()
        self.assertEqual(digid_metadata[:5], b"<?xml")

        entity_descriptor_node = etree.XML(digid_metadata)

        self.assertEqual(
            "http://test-entity.id", entity_descriptor_node.attrib["entityID"]
        )

        sspo_descriptor_node = entity_descriptor_node.find(
            ".//md:SPSSODescriptor",
            namespaces=NAME_SPACES,
        )

        self.assertEqual("true", sspo_descriptor_node.attrib["AuthnRequestsSigned"])
        self.assertEqual("true", sspo_descriptor_node.attrib["WantAssertionsSigned"])

        certificate_node = entity_descriptor_node.find(
            ".//ds:X509Certificate",
            namespaces=NAME_SPACES,
        )
        self.assertIn(
            "MIIC0DCCAbigAwIBAgIUEjGmfCGa1cOiTi+UKtDQVtySOHUwDQYJKoZIhvcNAQEL",
            certificate_node.text,
        )

        signature_algorithm_node = entity_descriptor_node.find(
            ".//ds:SignatureMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            signature_algorithm_node.attrib["Algorithm"],
        )

        digest_algorithm_node = entity_descriptor_node.find(
            ".//ds:DigestMethod",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://www.w3.org/2001/04/xmlenc#sha256",
            digest_algorithm_node.attrib["Algorithm"],
        )

        assertion_consuming_service_node = entity_descriptor_node.find(
            ".//md:AssertionConsumerService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(
            "http://test-entity.id/digid/acs/",
            assertion_consuming_service_node.attrib["Location"],
        )

        attribute_consuming_service_node = entity_descriptor_node.find(
            ".//md:AttributeConsumingService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("9050", attribute_consuming_service_node.attrib["index"])

        service_name_node = entity_descriptor_node.find(
            ".//md:ServiceName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Name", service_name_node.text)

        service_description_node = entity_descriptor_node.find(
            ".//md:ServiceDescription",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test Service Description", service_description_node.text)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_name_node.text)

        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("Test organisation", organisation_display_node.text)

        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("http://test-organisation.nl", organisation_url_node.text)

        contact_person_node = entity_descriptor_node.find(
            ".//md:ContactPerson",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("technical", contact_person_node.attrib["contactType"])

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("test@test.nl", contact_email_node.text)

        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )
        self.assertEqual("06123123123", contact_telephone_node.text)

        slo_nodes = entity_descriptor_node.findall(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertEqual(len(slo_nodes), 2)
        slo_soap, slo_redirect = slo_nodes
        self.assertEqual(
            slo_soap.attrib["Location"], "http://test-entity.id/digid/slo/soap/"
        )
        self.assertEqual(
            slo_soap.attrib["Binding"], "urn:oasis:names:tc:SAML:2.0:bindings:SOAP"
        )
        self.assertEqual(
            slo_redirect.attrib["Location"], "http://test-entity.id/digid/slo/redirect/"
        )
        self.assertEqual(
            slo_redirect.attrib["Binding"],
            "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
        )

    def test_contact_telephone_no_email(self):
        self.digid_config.technical_contact_person_telephone = "06123123123"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()
        entity_descriptor_node = etree.XML(digid_metadata)

        contact_email_node = entity_descriptor_node.find(
            ".//md:EmailAddress",
            namespaces=NAME_SPACES,
        )
        contact_telephone_node = entity_descriptor_node.find(
            ".//md:TelephoneNumber",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(contact_email_node)
        self.assertIsNone(contact_telephone_node)

    def test_organisation_url_no_service(self):
        self.digid_config.organization_url = "http://test-organisation.nl"
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()
        entity_descriptor_node = etree.XML(digid_metadata)

        organisation_name_node = entity_descriptor_node.find(
            ".//md:OrganizationName",
            namespaces=NAME_SPACES,
        )
        organisation_display_node = entity_descriptor_node.find(
            ".//md:OrganizationDisplayName",
            namespaces=NAME_SPACES,
        )
        organisation_url_node = entity_descriptor_node.find(
            ".//md:OrganizationURL",
            namespaces=NAME_SPACES,
        )

        self.assertIsNone(organisation_name_node)
        self.assertIsNone(organisation_display_node)
        self.assertIsNone(organisation_url_node)

    def test_slo_not_supported(self):
        self.digid_config.slo = False
        self.digid_config.save()

        digid_metadata = generate_digid_metadata()

        entity_descriptor_node = etree.XML(digid_metadata)

        single_logout_service_node = entity_descriptor_node.find(
            ".//md:SingleLogoutService",
            namespaces=NAME_SPACES,
        )
        self.assertIsNone(single_logout_service_node)


@pytest.mark.django_db
def test_current_and_next_certificate_in_metadata(
    temp_private_root,
    digid_config: DigidConfiguration,
    digid_certificate: Certificate,
    next_certificate: Certificate,
):
    ConfigCertificate.objects.create(
        config_type=ConfigTypes.digid,
        certificate=next_certificate,
    )
    assert ConfigCertificate.objects.count() == 2  # expect current and next

    digid_metadata = generate_digid_metadata()

    entity_descriptor_node = etree.XML(digid_metadata)

    metadata_node = entity_descriptor_node.find(
        "md:SPSSODescriptor", namespaces=NAME_SPACES
    )
    assert metadata_node is not None
    key_nodes = metadata_node.findall("md:KeyDescriptor", namespaces=NAME_SPACES)
    assert len(key_nodes) == 2  # we expect current + next key
    key1_node, key2_node = key_nodes
    assert key1_node.attrib["use"] == "signing"
    assert key2_node.attrib["use"] == "signing"

    with (
        digid_certificate.public_certificate.open("r") as _current,
        next_certificate.public_certificate.open("r") as _next,
    ):
        current_base64 = _current.read().replace("\n", "")
        next_base64 = _next.read().replace("\n", "")

    # certificate nodes include only the base64 encoded PEM data, without header/footer
    cert1_node = key1_node.find(
        "ds:KeyInfo/ds:X509Data/ds:X509Certificate", namespaces=NAME_SPACES
    )
    assert cert1_node is not None
    assert cert1_node.text is not None
    assert (cert_data_1 := cert1_node.text.strip()) in current_base64

    cert2_node = key2_node.find(
        "ds:KeyInfo/ds:X509Data/ds:X509Certificate", namespaces=NAME_SPACES
    )
    assert cert2_node is not None
    assert cert2_node.text is not None
    assert (cert_data_2 := cert2_node.text.strip()) in next_base64
    # they should not be the same
    assert cert_data_1 != cert_data_2



# Generated by Django 4.2.13 on 2024-07-19 10:53

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("simple_certmanager", "0002_alter_certificate_private_key_and_more"),
        ("digid_eherkenning", "0010_remove_digidconfiguration_key_passphrase_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="ConfigCertificate",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "config_type",
                    models.CharField(
                        choices=[
                            ("digid_eherkenning.DigidConfiguration", "DigiD"),
                            (
                                "digid_eherkenning.EherkenningConfiguration",
                                "eHerkenning",
                            ),
                        ],
                        max_length=100,
                        verbose_name="config type",
                    ),
                ),
                (
                    "certificate",
                    models.ForeignKey(
                        help_text="Certificate that may be used by the specified configuration. The best matching candidate will automatically be selected by the configuration.",
                        limit_choices_to={"type": "key_pair"},
                        on_delete=django.db.models.deletion.PROTECT,
                        to="simple_certmanager.certificate",
                        verbose_name="certificate",
                    ),
                ),
            ],
            options={
                "verbose_name": "DigiD/eHerkenning certificate",
                "verbose_name_plural": "DigiD/eHerkenning certificates",
            },
        ),
        migrations.AddConstraint(
            model_name="configcertificate",
            constraint=models.UniqueConstraint(
                fields=("config_type", "certificate"),
                name="uniq_config_cert",
                violation_error_message="This configuration and certificate combination already exists.",
            ),
        ),
    ]
